import pytest
import re
import numpy as np
import astropy.units as u
from astropy.coordinates import SkyCoord
from astropy.wcs import WCS
from astropy.tests.helper import assert_quantity_allclose


def assert_coordinate_close(coord1, coord2, atol=1 * u.arcsec):
    # check that two coordinates are within some separation tolerance
    separation = coord1.separation(coord2)
    assert_quantity_allclose(separation, desired=0*u.arcsec, atol=atol)


def assert_angle_close(angle1, angle2, atol=1 * u.arcsec):
    # check that two angles are within some separation tolerance
    difference = abs(angle1 - angle2)
    assert_quantity_allclose(difference, desired=0*u.arcsec, atol=atol)


def test_get_viewport_sky(imviz_helper, image_hdu_wcs):
    imviz_helper.load_data(image_hdu_wcs)
    imviz_helper.plugins['Orientation'].align_by = 'WCS'

    viewer = imviz_helper.app.get_viewer('imviz-0')
    viewport = viewer.aid.get_viewport('sky')

    expected_center = SkyCoord(ra=337.51894337, dec=-20.83208305, unit='deg')
    expected_fov = 0.00277778 * u.deg
    expected_image_label = imviz_helper.app.data_collection[0].label

    # by default, the viewer y-axis is the narrower axis, which is used to
    # define the FOV parameter. Use the WCS to find what the actual
    # viewport dimensions are in the x-axis, which maps onto RA in this case:
    wcs = WCS(image_hdu_wcs.header)
    dec_unit = u.Unit(wcs.wcs.cunit[1])
    delta_dec = abs(wcs.wcs.cdelt[1]) * dec_unit
    expected_fov = (
        abs(viewer.state.y_max - viewer.state.y_min) * delta_dec
    )
    assert_coordinate_close(viewport['center'], expected_center)
    assert_angle_close(viewport['fov'], expected_fov)

    expected_rotation = 360 * u.deg
    assert_angle_close(viewport['rotation'], expected_rotation, atol=0.001*u.deg)

    assert viewport['image_label'] == expected_image_label


def test_set_viewport_sky(imviz_helper, image_hdu_wcs):
    imviz_helper.load_data(image_hdu_wcs)
    imviz_helper.plugins['Orientation'].align_by = 'WCS'
    viewer = imviz_helper.app.get_viewer('imviz-0')

    # change only the center:
    new_viewport_settings = dict(
        center=SkyCoord(ra=337.5, dec=-20.8, unit='deg'),
        fov=0.01 * u.deg,
    )
    viewer.aid.set_viewport(**new_viewport_settings)
    new_viewport = viewer.aid.get_viewport('sky')

    assert_coordinate_close(new_viewport['center'], new_viewport_settings['center'])

    # todo: investigate why this tolerance needs to be larger than expected:
    assert_angle_close(new_viewport['fov'], new_viewport_settings['fov'], atol=1 * u.arcsec)

    with pytest.raises(ValueError, match='The AID API supports `center` arguments as'):
        viewer.aid.set_viewport(center=u.Quantity([0, 1]))


def test_set_viewport_sky_rotation(imviz_helper, image_hdu_wcs):
    imviz_helper.load_data(image_hdu_wcs)
    imviz_helper.plugins['Orientation'].align_by = 'WCS'
    viewer = imviz_helper.app.get_viewer('imviz-0')

    # change only the center:
    new_viewport_settings = dict(
        rotation=90 * u.deg
    )
    viewer.aid.set_viewport(**new_viewport_settings)
    new_viewport = viewer.aid.get_viewport('sky')

    assert_angle_close(
        new_viewport['rotation'],
        new_viewport_settings['rotation'],
        atol=0.001*u.deg
    )


def test_get_viewport_external_update(imviz_helper, image_hdu_wcs):
    # arrange
    imviz_helper.load_data(image_hdu_wcs)
    imviz_helper.plugins['Orientation'].align_by = 'WCS'
    viewer = imviz_helper.app.get_viewer('imviz-0')

    # act: update the rotation of imviz external to aida
    orientation = imviz_helper.app._jdaviz_helper.plugins.get('Orientation', None)
    orientation.add_orientation(
        east_left=True,
        set_on_create=True,
        label='test',
        rotation_angle=90.0
    )

    # assert: check that aida get viewport gets the correct rotation
    new_viewport = viewer.aid.get_viewport('sky')
    assert_angle_close(
        new_viewport['rotation'],
        90.0*u.deg,
        atol=0.001*u.deg
    )


def test_set_viewport_pixel(imviz_helper, image_hdu_wcs):
    imviz_helper.load_data(image_hdu_wcs)
    imviz_helper.plugins['Orientation'].align_by = 'WCS'

    viewer = imviz_helper.app.get_viewer('imviz-0')

    # change only the center:
    new_viewport_settings = dict(
        center=[5, 5],
        fov=10
    )
    viewer.aid.set_viewport(**new_viewport_settings)
    new_viewport = viewer.aid.get_viewport('pixel')

    np.testing.assert_allclose(new_viewport['center'], new_viewport_settings['center'])

    # todo: investigate why this tolerance needs to be larger than expected:
    np.testing.assert_allclose(new_viewport['fov'], new_viewport_settings['fov'], atol=1e-4)


def test_set_rotation_not_aligned(imviz_helper, image_hdu_wcs):
    imviz_helper.load_data(image_hdu_wcs)
    viewer = imviz_helper.app.get_viewer('imviz-0')

    # change only the center:
    new_viewport_settings = dict(
        rotation=90 * u.deg
    )
    with pytest.raises(ValueError, match=re.escape("The viewer must be aligned by WCS to use `set_rotation`.")):  # noqa
        viewer.aid.set_viewport(**new_viewport_settings)


def test_get_fov_no_wcs(imviz_helper, image_hdu_nowcs):
    imviz_helper.load_data(image_hdu_nowcs)
    viewer = imviz_helper.app.get_viewer('imviz-0')

    with pytest.raises(ValueError, match=re.escape("The image must have valid WCS to return `fov` in `sky`.")):  # noqa
        viewer.aid.get_viewport()



import pytest
from jdaviz.core.user_api import ViewerCreatorUserApi


def test_viewer_creator_relevance(deconfigged_helper, spectrum1d, spectrum2d):
    assert len(deconfigged_helper.new_viewers) == 0
    deconfigged_helper.load(spectrum1d, format='1D Spectrum')
    # 1D Spectrum, Scatter, Histogram
    assert len(deconfigged_helper.new_viewers) == 3
    deconfigged_helper.load(spectrum2d, format='2D Spectrum')
    # 1D Spectrum, 2D Spectrum, Scatter, Histogram
    assert len(deconfigged_helper.new_viewers) == 4


def test_viewer_creation(deconfigged_helper, spectrum1d):
    deconfigged_helper.load(spectrum1d, format='1D Spectrum')
    assert len(deconfigged_helper.viewers) == 1
    deconfigged_helper.new_viewers['1D Spectrum']()
    assert len(deconfigged_helper.viewers) == 2


class TestViewerCreatorObject:
    @pytest.fixture(autouse=True)
    def _setup(self, deconfigged_helper, spectrum1d, spectrum1d_nm, spectrum2d):
        self.spectrum1d = spectrum1d
        self.spectrum1d_nm = spectrum1d_nm
        self.spectrum2d = spectrum2d
        deconfigged_helper.load(spectrum1d, format='1D Spectrum')
        self.dcf_helper = deconfigged_helper
        self.creator = deconfigged_helper.new_viewers['1D Spectrum']._obj

    def test_viewer_creator_initialization(self):
        """
        Test that viewer creator initializes with correct defaults.
        """
        assert self.creator.viewer_type == '1D Spectrum'
        assert self.creator.is_relevant is True
        assert self.creator.viewer_label_auto is True
        # Since we autouse the fixture, there is already one loaded dataset
        # so the default now has (1)
        assert self.creator.viewer_label_default == '1D Spectrum (1)'
        assert self.creator.viewer_label_invalid_msg == ''
        assert self.creator.dataset.multiselect is True

        # Check that callbacks exist (they may be None or callable)
        assert hasattr(self.creator, 'set_active_callback')
        assert hasattr(self.creator, 'open_callback')
        assert hasattr(self.creator, 'close_callback')

        # Check that user_api property returns the correct type
        user_api = self.dcf_helper.new_viewers['1D Spectrum']
        assert isinstance(user_api, ViewerCreatorUserApi)

        assert 'is_spectrum' in self.creator.dataset.filters

    def test_viewer_creator_close_in_tray(self):
        """
        Test close_in_tray method with and without closing sidebar.
        """
        # Test without closing sidebar
        self.creator.close_in_tray(close_sidebar=False)
        # Should not raise an error even if callback is None

        # Test with closing sidebar
        self.dcf_helper.app.state.drawer_content = 'test_content'
        self.creator.close_in_tray(close_sidebar=True)
        assert self.dcf_helper.app.state.drawer_content == ''

    def test_viewer_creator_open_in_tray_callbacks(self):
        """
        Test that open_in_tray raises error without set_active_callback
        and succeeds with proper callbacks.
        """
        # Remove the callback to test error handling
        self.creator.set_active_callback = None

        msg = 'set_active_callback must be set to open dialog to specific tab'
        with pytest.raises(NotImplementedError, match=msg):
            self.creator.open_in_tray()

        # Now test with proper callbacks
        callback_called = []

        def mock_set_active(label):
            callback_called.append(('set_active', label))

        def mock_open():
            callback_called.append(('open',))

        self.creator.set_active_callback = mock_set_active
        self.creator.open_callback = mock_open

        self.creator.open_in_tray()

        assert ('set_active', '1D Spectrum') in callback_called
        assert ('open',) in callback_called

    def test_viewer_creator_is_relevant_changes(self):
        """
        Test that is_relevant updates based on available datasets.
        """
        assert self.creator.is_relevant is True

        # Check that it appears in new_viewer_items
        labels = [ti['label'] for ti in self.dcf_helper.app.state.new_viewer_items]
        assert '1D Spectrum' in labels
        idx = labels.index('1D Spectrum')
        assert self.dcf_helper.app.state.new_viewer_items[idx]['is_relevant'] is True

    def test_viewer_label_validation_duplicate(self):
        """
        Test that duplicate viewer labels are properly rejected.
        """
        # Create first viewer
        viewer1 = self.dcf_helper.new_viewers['1D Spectrum']()
        assert viewer1._obj.id == '1D Spectrum (1)'

        # Try to create another with the same label
        creator = self.dcf_helper.new_viewers['1D Spectrum']._obj
        # The default should update to avoid conflict
        assert creator.viewer_label_default != '1D Spectrum'

        creator.viewer_label_value = '1D Spectrum'

        msg = "Viewer label '1D Spectrum' already in use."
        assert creator.viewer_label_invalid_msg == msg

        # Calling should raise ValueError
        with pytest.raises(ValueError, match=msg):
            creator()

    def test_viewer_creator_with_datasets(self):
        """
        Test creating viewer with single/multiple selected datasets.
        """
        # Select dataset
        self.creator.dataset.selected = ['1D Spectrum']

        # Create viewer
        viewer = self.creator()
        # Viewer should have the dataset loaded
        assert len(viewer._obj.data_collection) == 1

        # Load in a second spectrum but first check that the observer works
        assert self.creator.is_relevant is True
        initial_item_count = len(self.creator.dataset.items)
        # Load another one in
        self.dcf_helper.load(self.spectrum1d_nm, format='1D Spectrum')

        # Should still be relevant with more items per the observer
        assert self.creator.is_relevant is True
        assert len(self.creator.dataset.items) > initial_item_count

        # Select multiple datasets
        available_datasets = [item['label'] for item in self.creator.dataset.items]
        self.creator.dataset.selected = available_datasets[:2]
        viewer = self.creator()

        # Viewer should have both datasets
        assert len(viewer._obj.data_collection) == 2

    def test_viewer_creator_observer_without_viewer_attribute(self):
        """
        Test that observers handle missing viewer attribute gracefully.
        """
        # Temporarily remove viewer attribute if it exists
        had_viewer = hasattr(self.creator, 'viewer')
        original_viewer = None
        if had_viewer:
            original_viewer = self.creator.viewer
            delattr(self.creator, 'viewer')

        # These should not raise errors
        self.creator._viewer_label_value_changed()
        self.creator._viewer_items_changed()

        # Restore viewer attribute
        if had_viewer and original_viewer is not None:
            self.creator.viewer = original_viewer

    def test_multiple_viewer_types(self):
        """
        Test that multiple viewer types can coexist.
        """
        self.dcf_helper.load(self.spectrum2d, format='2D Spectrum')

        # Should have multiple viewer types available
        assert '1D Spectrum' in self.dcf_helper.new_viewers
        assert '2D Spectrum' in self.dcf_helper.new_viewers

        # Create viewers of different types
        viewer1d = self.dcf_helper.new_viewers['1D Spectrum']()
        viewer2d = self.dcf_helper.new_viewers['2D Spectrum']()

        assert viewer1d._obj.id != viewer2d._obj.id

    def test_vue_create_clicked(self):
        """
        Test that vue_create_clicked calls the viewer creator.
        """
        initial_viewer_count = len(self.dcf_helper.viewers)

        # Call vue method
        self.creator.vue_create_clicked()

        # Should have created a new viewer
        assert len(self.dcf_helper.viewers) == initial_viewer_count + 1

    def test_viewer_creator_with_custom_label(self):
        """
        Test creating viewer with custom label.
        """
        # Set custom label
        custom_label = 'My Custom Viewer'
        self.creator.viewer_label_value = custom_label
        self.creator.viewer_label_auto = False

        # Should not have validation error
        assert self.creator.viewer_label_invalid_msg == ''

        # Create viewer
        viewer = self.creator()

        # Should have custom label
        assert viewer._obj.id == custom_label



from jdaviz.configs.default.plugins.viewers import JdavizTableViewer
from jdaviz.core.user_api import ViewerCreatorUserApi
from jdaviz.core.viewer_creators import BaseViewerCreator
from jdaviz.core.registries import viewer_creator_registry


__all__ = ['TableViewerCreator']


@viewer_creator_registry('Table', overwrite=True)
class TableViewerCreator(BaseViewerCreator):
    template_file = __file__, "table.vue"

    def __init__(self, app, **kwargs):
        super().__init__(app, **kwargs)
        self.dataset.filters = ['is_catalog']
        self.viewer_label_default = 'Table'

    @property
    def user_api(self):
        return ViewerCreatorUserApi(self)

    @property
    def viewer_class(self):
        return JdavizTableViewer



import numpy as np
from numpy.testing import assert_allclose
import pytest

import astropy.units as u
from astropy.table import QTable
from specutils import Spectrum

from jdaviz.core.marks import SpectralLine
from jdaviz.core.linelists import get_available_linelists


# two-argument Table.loc is deprecated as of Astropy 7.2. Syntax update will be needed
# and is as below:
# Table.loc_indices.with_index(...):
# https://docs.astropy.org/en/latest/whatsnew/7.2.html
@pytest.mark.filterwarnings('ignore:Calling `Table\\.loc/iloc/loc_indices\\[index_id, item\\]`.*:astropy.utils.exceptions.AstropyDeprecationWarning')  # noqa
class TestLineLists:
    def test_line_lists(self, specviz_helper):
        spec = Spectrum(flux=np.random.rand(100)*u.Jy,
                        spectral_axis=np.arange(6000, 7000, 10)*u.AA)
        specviz_helper.load_data(spec)

        lt = QTable()
        lt['linename'] = ['O III', 'Halpha']
        lt['rest'] = [0, 6563]*u.AA
        with pytest.raises(ValueError, match='all rest values must be positive'):
            specviz_helper.load_line_list(lt)

        lt = QTable()
        lt['linename'] = ['O III', 'Halpha']
        lt['rest'] = [5007, 6563]*u.AA
        specviz_helper.load_line_list(lt)

        assert len(specviz_helper.spectral_lines) == 2
        assert specviz_helper.spectral_lines.loc["linename", "Halpha"]["listname"] == "Custom"
        assert np.all(specviz_helper.spectral_lines["show"])
        assert specviz_helper.plugins['Line Lists']._obj.rs_enabled is True

        specviz_helper.erase_spectral_lines()
        assert np.all(specviz_helper.spectral_lines["show"] == False)  # noqa
        assert specviz_helper.plugins['Line Lists']._obj.rs_enabled is False

        specviz_helper.plot_spectral_line("Halpha")
        specviz_helper.plot_spectral_line("O III 5007.0")

        assert np.all(specviz_helper.spectral_lines["show"])

        assert (
            specviz_helper.plugins["Line Lists"]._obj.list_contents["Custom"]["medium"]
            == "Unknown (Custom)"
        )

    def test_redshift(self, specviz_helper, spectrum1d):
        # Also test that plugin is disabled before data is loaded.
        ll_plugin = specviz_helper.plugins['Line Lists']._obj
        assert ll_plugin.disabled_msg

        label = "Test 1D Spectrum"
        specviz_helper.load_data(spectrum1d, data_label=label)

        assert not ll_plugin.disabled_msg

        lt = QTable()
        lt['linename'] = ['O III', 'Halpha']
        lt['rest'] = [5007, 6563]*u.AA
        lt['redshift'] = u.Quantity(0.046)
        lt['listname'] = 'Test List'
        with pytest.warns(UserWarning, match='per line/list redshifts not supported, use viz.set_redshift'):  # noqa
            specviz_helper.load_line_list(lt)

        # open the plugin so that all updates run
        ll_plugin.plugin_opened = True
        line = ll_plugin.list_contents['Test List']['lines'][0]
        assert_allclose(line['obs'], line['rest'])
        # test API access
        specviz_helper.set_redshift(0.01)
        specviz_helper.set_redshift_slider_bounds(range=0.5, step=0.01)
        specviz_helper.set_redshift_slider_bounds(range='auto', step='auto')

        # test plugin
        ll_plugin.rs_redshift = 0.1
        assert ll_plugin.rs_rv == 28487.06614479641

        ll_plugin.rs_rv = 30000
        assert ll_plugin.rs_redshift == 0.10561890816244568

        # https://github.com/spacetelescope/jdaviz/issues/1692
        # adding new data entry from a plugin should not reset redshift
        specviz_helper.plugins['Gaussian Smooth'].smooth()
        assert ll_plugin.rs_redshift == 0.10561890816244568

        # test that setting observed wavelength works
        ll_plugin.vue_change_line_obs({'list_name': 'Test List',
                                       'line_ind': 0,
                                       'obs_new': 5508})
        assert_allclose(line['obs'], 5508)
        assert ll_plugin.rs_redshift == 0.10005991611743559

        # https://github.com/spacetelescope/jdaviz/issues/1168
        ll_plugin.vue_set_identify(('Test List', line, 0))
        ll_plugin.vue_remove_list('Test List')
        assert ll_plugin.spectrum_viewer.spectral_lines is None
        assert ll_plugin.identify_label == ''

    def test_load_available_preset_lists(self, specviz_helper, spectrum1d):
        """ Loads all available line lists and checks the medium requirement """
        label = "Test 1D Spectrum"
        specviz_helper.load_data(spectrum1d, data_label=label)

        # Check to make sure we got our line lists
        available_linelists = get_available_linelists()
        assert len(available_linelists) > 0

        for linelist in available_linelists:
            specviz_helper.plugins['Line Lists']._obj.vue_list_selected(linelist)
            specviz_helper.plugins['Line Lists']._obj.vue_load_list(linelist)

        # Check that we loaded all the lists (+1 because of the Custom list)
        assert (
            len(specviz_helper.plugins['Line Lists']._obj.list_contents.keys()) ==
            len(available_linelists) + 1
        )

        # Line list must have "medium" info to be available
        for list in specviz_helper.plugins['Line Lists']._obj.list_contents.values():  # noqa
            assert 'medium' in list

    def test_line_identify(self, specviz_helper, spectrum1d):
        specviz_helper.load_data(spectrum1d)

        lt = QTable()
        lt['linename'] = ['O III', 'Halpha']
        lt['rest'] = [5007, 6563]*u.AA
        lt['listname'] = 'Test List'
        specviz_helper.load_line_list(lt)

        ll_plugin = specviz_helper.app.get_tray_item_from_name('g-line-list')
        line = ll_plugin.list_contents['Test List']['lines'][0]
        assert line.get('identify', False) is False
        ll_plugin.vue_set_identify(('Test List', line, 0))
        assert line.get('identify', False) is True

        ll_plugin.vue_change_visible(('Test List', line, 0))
        assert line.get('show') is False
        assert line.get('identify', False) is False

    def test_global_redshift_applied(self, specviz_helper, spectrum1d):
        specviz_helper.load_data(spectrum1d)

        # Create a table with redshift included
        lt = QTable({'linename': ['O III', 'Halpha'],
                     'rest': [5007, 6563] * u.AA,
                     'redshift': u.Quantity([0, 0])})

        with pytest.warns(UserWarning, match='per line/list redshifts not supported, use viz.set_redshift'):  # noqa
            specviz_helper.load_line_list(lt)

        # Load a line, and apply redshift globally
        specviz_helper.plot_spectral_line("Halpha")
        specviz_helper.set_redshift(0.01)
        # Load second line, redshift should also be applied to it
        specviz_helper.plot_spectral_line("O III")

        viewer_lines = [mark for mark in specviz_helper.app.get_viewer(
            specviz_helper._default_spectrum_viewer_reference_name).figure.marks
            if isinstance(mark, SpectralLine)]

        assert np.allclose([line.redshift for line in viewer_lines], 0.01)

    def test_global_redshift_applied_to_all(self, specviz_helper, spectrum1d):
        specviz_helper.load_data(spectrum1d)

        # Create a table with redshift included
        lt = QTable({'linename': ['O III', 'Halpha', 'O I'],
                     'rest': [5007, 6563, 6300] * u.AA,
                     'redshift': u.Quantity([0, 0, 0])})

        with pytest.warns(UserWarning, match='per line/list redshifts not supported, use viz.set_redshift'):  # noqa
            specviz_helper.load_line_list(lt)

        # Load a line, so we can apply redshift
        specviz_helper.plot_spectral_line("Halpha")
        global_redshift = 0.01
        specviz_helper.set_redshift(global_redshift)
        # Load remaining lines
        specviz_helper.plot_spectral_lines(global_redshift)

        viewer_lines = [mark for mark in specviz_helper.app.get_viewer(
            specviz_helper._default_spectrum_viewer_reference_name).figure.marks
            if isinstance(mark, SpectralLine)]

        assert np.allclose([line.redshift for line in viewer_lines], 0.01)



import numpy as np
from astropy import units as u
from astropy.coordinates import SkyCoord
from traitlets import Bool, Unicode, Instance, observe, default
import ipywidgets as widgets

from jdaviz.core.events import (ViewerAddedMessage, ChangeRefDataMessage,
                                AddDataMessage, RemoveDataMessage, MarkersPluginUpdate)
from jdaviz.core.marks import MarkersMark, DistanceMeasurement
from jdaviz.core.registries import tray_registry
from jdaviz.core.template_mixin import (PluginTemplateMixin, ViewerSelectMixin, TableMixin,
                                        Table, _is_image_viewer)
from jdaviz.core.user_api import PluginUserApi


__all__ = ['Markers']


@tray_registry('g-markers', label="Markers",
               category='core', sidebar='info', subtab=1)
class Markers(PluginTemplateMixin, ViewerSelectMixin, TableMixin):
    """
    See the :ref:`Markers Plugin Documentation <markers-plugin>` for more details.

    Only the following attributes and methods are available through the
    :ref:`public plugin API <plugin-apis>`:

    * :meth:`~jdaviz.core.template_mixin.PluginTemplateMixin.show`
    * :meth:`~jdaviz.core.template_mixin.PluginTemplateMixin.open_in_tray`
    * :meth:`~jdaviz.core.template_mixin.PluginTemplateMixin.close_in_tray`
    * :meth:`~jdaviz.core.template_mixin.TableMixin.clear_table`
    * :meth:`~jdaviz.core.template_mixin.TableMixin.export_table`
    """
    template_file = __file__, "markers.vue"
    uses_active_status = Bool(True).tag(sync=True)
    distance_display = Unicode("N/A").tag(sync=True)
    measurements_table = Instance(Table).tag(sync=True, **widgets.widget_serialization)

    _default_table_values = {'spectral_axis': np.nan,
                             'spectral_axis:unit': '',
                             'slice': np.nan,
                             'pixel_x': np.nan,
                             'pixel_y': np.nan,
                             'pixel:unreliable': None,
                             'world_ra': np.nan,
                             'world_dec': np.nan,
                             'world:unreliable': None,
                             'value': np.nan,
                             'value:unit': '',
                             'value:unreliable': None,
                             'index': np.nan}

    @property
    def user_api(self):
        return PluginUserApi(self, expose=('clear_table', 'export_table',))

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if self.config == 'cubeviz':
            headers = ['spectral_axis', 'spectral_axis:unit',
                       'slice', 'pixel_x', 'pixel_y',
                       'world_ra', 'world_dec', 'value', 'value:unit', 'viewer']

        elif self.config == 'imviz':
            headers = ['pixel_x', 'pixel_y', 'pixel:unreliable',
                       'world_ra', 'world_dec', 'world:unreliable',
                       'value', 'value:unit', 'value:unreliable',
                       'viewer']

        elif self.config == 'specviz':
            headers = ['spectral_axis', 'spectral_axis:unit',
                       'index', 'value', 'value:unit', 'viewer']

        elif self.config in ('specviz2d'):
            # TODO: add "index" if/when specviz2d supports plotting spectral_axis
            headers = ['spectral_axis', 'spectral_axis:unit',
                       'pixel_x', 'pixel_y', 'value', 'value:unit', 'viewer']

        elif self.config == 'mosviz':
            headers = ['spectral_axis', 'spectral_axis:unit',
                       'pixel_x', 'pixel_y', 'world_ra', 'world_dec', 'index',
                       'value', 'value:unit', 'viewer']
        elif self.config == 'deconfigged':
            # for now combination of specviz2d + imviz (will eventually need more)
            headers = ['spectral_axis', 'spectral_axis:unit',
                       'pixel_x', 'pixel_y', 'pixel:unreliable',
                       'world_ra', 'world_dec', 'world:unreliable',
                       'value', 'value:unit', 'value:unreliable',
                       'viewer']
        else:
            # allow downstream configs to override headers
            headers = kwargs.get('headers', [])

        headers += ['data_label']
        self.table.headers_avail = headers
        self.table.headers_visible = headers
        self.table._default_values_by_colname = self._default_table_values

        self._distance_marks = {}
        self._distance_first_point = None
        self._temporary_dist_measure = None
        self._viewer_for_drawing = None

        # Set the callbacks for each table's clear button
        self.table._clear_callback = self._clear_markers_table_callback
        self.measurements_table._clear_callback = self._clear_measurements_table_callback

        # subscribe to mouse events on any new viewers
        self.hub.subscribe(self, ViewerAddedMessage, handler=self._on_viewer_added)

        # account for image rotation due to a change in reference data
        # and reset distance tool if orientation changes mid-measurement
        self.hub.subscribe(self, ChangeRefDataMessage,
                           handler=self._on_alignment_change)

        # enable/disable mark based on whether parent data entry is in viewer
        self.hub.subscribe(self, AddDataMessage,
                           handler=lambda msg: self._recompute_mark_positions(msg.viewer))
        self.hub.subscribe(self, RemoveDataMessage,
                           handler=lambda msg: self._recompute_mark_positions(msg.viewer))

        self.docs_description = """Press 'm' with the cursor over a viewer to log
                                 the mouseover information. To change the
                                 selected layer, click the layer cycler in the
                                 mouseover information section of the app-level
                                 toolbar.
                                 Press 'd' twice to measure the distance between any two points.
                                 Hold the 'Option'(Mac) or 'Alt' key while pressing 'd' to snap
                                 measurement points to the nearest marker."""
        # description displayed under plugin title in tray
        self._plugin_description = 'Create markers on viewers.'

    @default('measurements_table')
    def _default_measurements_table(self):
        table = Table(
            self,
            name='measurements_table',
            title='Measurements',
            headers=['Start RA', 'Start Dec', 'End RA', 'End Dec',
                     'Separation (arcsec)', 'Distance (pix)', 'Position Angle (deg)',
                     'Start X', 'End X', 'Start Y', 'End Y',
                     'Δx', 'Δy', 'Viewer', 'Data Label'],
        )
        return table

    def _on_alignment_change(self, msg=None):
        """
        Resets the distance measurement tool if the orientation changes
        while a measurement is in progress.
        """
        if self._distance_first_point is not None:
            # If we are in the middle of drawing, clean up the temporary marks
            if self._temporary_dist_measure is not None and self._viewer_for_drawing is not None:
                self._viewer_for_drawing.remove_event_callback(self._on_mouse_move_while_drawing)
                temp_marks = self._temporary_dist_measure.marks
                self._viewer_for_drawing.figure.marks = [
                    m for m in self._viewer_for_drawing.figure.marks if m not in temp_marks
                ]
                self._temporary_dist_measure = None
                self._viewer_for_drawing = None

            self._distance_first_point = None
            self.distance_display = "N/A"
        if msg is not None:
            self._recompute_mark_positions(msg.viewer)

    def _clear_markers_table_callback(self, *args):
        """
        Callback for the clear button on the main markers table.
        This clears all 'm' key marks from all viewers.
        """
        for mark in self.marks.values():
            mark.clear()

        # The only job here is to announce that the table is empty.
        self.hub.broadcast(MarkersPluginUpdate(table_length=0, sender=self))

    def _clear_measurements_table_callback(self, *args):
        """
        Callback for the clear button on the measurements table.
        This clears all 'd' key distance lines from all viewers.
        """
        all_dist_marks = set()
        for dist_measures in self._distance_marks.values():
            for dm in dist_measures:
                all_dist_marks.update(dm.marks)

        # Iterate through all viewers in the app and remove any distance marks found.
        for viewer in self.app._viewer_store.values():
            if not hasattr(viewer, 'figure'):
                continue
            viewer.figure.marks = [m for m in viewer.figure.marks if m not in all_dist_marks]
        self._distance_marks.clear()
        self.distance_display = "N/A"
        self._distance_first_point = None

    @observe('table.items')
    def _sync_markers_from_table(self, *args):
        """
        Observer callback that redraws all marks when the markers table changes
        (e.g., a row is deleted) to ensure the plot and table are synchronized.
        """
        for viewer in self.app._viewer_store.values():
            if hasattr(viewer, 'figure'):
                self._recompute_mark_positions(viewer)

    def _create_viewer_callbacks(self, viewer):
        if not self.is_active:
            return
        callback = self._viewer_callback(viewer, self._on_viewer_key_event)
        viewer.add_event_callback(callback, events=['keydown'])

    def _on_viewer_added(self, msg):
        self._create_viewer_callbacks(self.app.get_viewer_by_id(msg.viewer_id))

    def _recompute_mark_positions(self, viewer):
        if self.table is None or self.table._qtable is None:
            return
        if 'world_ra' not in self.table.headers_avail:
            return
        viewer_id = viewer.reference if viewer.reference is not None else viewer.reference_id
        viewer_loaded_data = [lyr.layer.label for lyr in viewer.layers]
        data_labels = self.table._qtable['data_label']
        viewer_labels = self.table._qtable['viewer']
        # note: could eventually have a user-provided switch to show markers in other viewers
        # by just skipping this first viewer_label == viewer_id check
        in_viewer = [viewer_label == viewer_id and data_label in viewer_loaded_data
                     for viewer_label, data_label in zip(viewer_labels, data_labels)]
        viewer_mark = self._get_mark(viewer)
        if not np.any(in_viewer):
            viewer_mark.x, viewer_mark.y = [], []
            return

        if self.app._align_by.lower() == 'wcs':
            # convert from the sky coordinates in the table to pixels via the WCS of the current
            # reference data
            orig_world_x = np.asarray(self.table._qtable['world_ra'][in_viewer])
            orig_world_y = np.asarray(self.table._qtable['world_dec'][in_viewer])
            pixel_unreliable = np.asarray(self.table._qtable['pixel:unreliable'][in_viewer])
            new_wcs = viewer.state.reference_data.coords
            try:
                new_x, new_y = new_wcs.world_to_pixel_values(orig_world_x*u.deg,
                                                             orig_world_y*u.deg)
                for coord in [new_x, new_y]:
                    coord[pixel_unreliable] = np.nan
            except Exception:
                # fail gracefully
                new_x, new_y = [], []

        elif self.app._align_by.lower() == 'pixels':
            # When aligning by pixels, we can just use the pixel coordinates
            # that were stored in the table when the marks were created.
            new_x = np.asarray(self.table._qtable['pixel_x'][in_viewer])
            new_y = np.asarray(self.table._qtable['pixel_y'][in_viewer])

        else:
            raise NotImplementedError(f"align_by {self.app._align_by} not implemented")

        # check for entries that do not correspond to a layer or only have pixel coordinates
        pixel_only_inds = data_labels == ''
        if np.any(pixel_only_inds):
            # TODO: should we rescale these since pixel coordinates when aligned by WCS are always
            # on the range 0-1 because of the orientation layer? Or hide the pixel option in the
            # cycler when WCS-linked?
            pixel_x = np.asarray(self.table._qtable['pixel_x'])
            pixel_y = np.asarray(self.table._qtable['pixel_y'])
            new_x = np.append(new_x, pixel_x[pixel_only_inds])
            new_y = np.append(new_y, pixel_y[pixel_only_inds])

        viewer_mark.x, viewer_mark.y = new_x, new_y

    def _add_sky_distance_row(self, p1, p2, viewer):
        world_avail = ('world_ra' in p1 and 'world_ra' in p2 and
                       p1.get('world_ra') is not None and p2.get('world_ra') is not None and
                       np.all(np.isfinite([p1['world_ra'], p1['world_dec'],
                                           p2['world_ra'], p2['world_dec']])))
        if world_avail:
            c1 = SkyCoord(p1['world_ra'], p1['world_dec'], unit='deg', frame='icrs')
            c2 = SkyCoord(p2['world_ra'], p2['world_dec'], unit='deg', frame='icrs')
            sep_arcsec = c1.separation(c2).arcsec
            pos_angle = c1.position_angle(c2).deg
            start_ra, start_dec = p1['world_ra'], p1['world_dec']
            end_ra, end_dec = p2['world_ra'], p2['world_dec']
        else:
            sep_arcsec, pos_angle, start_ra, start_dec, end_ra, end_dec = (np.nan,) * 6

        dist_pix = np.sqrt((p2.get('pixel_x', 0) - p1.get('pixel_x', 0))**2 +
                           (p2.get('pixel_y', 0) - p1.get('pixel_y', 0))**2)
        viewer_id = viewer.reference or viewer.reference_id
        data_label = p1.get('data_label', 'N/A')

        row_data = {
            'Start RA': f"{start_ra:.4f}" if not np.isnan(start_ra) else "N/A",
            'Start Dec': f"{start_dec:.4f}" if not np.isnan(start_dec) else "N/A",
            'End RA': f"{end_ra:.4f}" if not np.isnan(end_ra) else "N/A",
            'End Dec': f"{end_dec:.4f}" if not np.isnan(end_dec) else "N/A",
            'Separation (arcsec)': f"{sep_arcsec:.4f}" if not np.isnan(sep_arcsec) else "N/A",
            'Distance (pix)': f"{dist_pix:.2f}",
            'Position Angle (deg)': f"{pos_angle:.2f}" if not np.isnan(pos_angle) else "N/A",
            'Viewer': viewer_id,
            'Data Label': data_label
        }
        self.measurements_table.add_item(row_data)

    def _add_profile_distance_row(self, p1, p2, viewer, dx, dy, x_unit, y_unit):
        """Adds a single row to the measurements table for a profile-based distance."""
        viewer_id = viewer.reference or viewer.reference_id
        data_label = p1.get('data_label', 'N/A')

        # Get start/end coordinates from the point dictionaries
        start_x = p1.get('axes_x', 0)
        end_x = p2.get('axes_x', 0)
        start_y = p1.get('axes_y', 0)
        end_y = p2.get('axes_y', 0)

        row_data = {
            'Start X': f"{start_x:.4g} {x_unit}",
            'End X': f"{end_x:.4g} {x_unit}",
            'Start Y': f"{start_y:.4g} {y_unit}",
            'End Y': f"{end_y:.4g} {y_unit}",
            'Δx': f"{dx:.4g} {x_unit}",
            'Δy': f"{dy:.4g} {y_unit}",
            'Viewer': viewer_id,
            'Data Label': data_label
        }
        self.measurements_table.add_item(row_data)

    def _calculate_distance_text(self, p1, p2, viewer, add_row=False):
        """
        Helper function to calculate the distance between two points and
        generate the appropriate display text. Optionally adds a row to the
        measurements table.
        """
        is_profile = not _is_image_viewer(viewer)

        if is_profile:
            dx_val = p2.get('axes_x', 0) - p1.get('axes_x', 0)
            dy_val = p2.get('axes_y', 0) - p1.get('axes_y', 0)
            x_unit = p1.get('axes_x:unit', '')
            y_unit = p1.get('axes_y:unit', '')
            text_ui = f"Δx: {dx_val:.4g} {x_unit}, Δy: {dy_val:.4g} {y_unit}"
            text_plot = ""
            if add_row:
                self._add_profile_distance_row(p1, p2, viewer, dx_val, dy_val, x_unit, y_unit)
        else:  # Assume image viewer
            world_avail = ('world_ra' in p1 and 'world_ra' in p2 and
                           p1.get('world_ra') is not None and p2.get('world_ra') is not None and
                           np.all(np.isfinite([p1['world_ra'], p1['world_dec'],
                                               p2['world_ra'], p2['world_dec']])))
            if self.config == 'imviz' and self.app._align_by.lower() == 'pixels':
                dist_pix = np.sqrt((p2.get('pixel_x', 0) - p1.get('pixel_x', 0))**2 +
                                   (p2.get('pixel_y', 0) - p1.get('pixel_y', 0))**2)
                text_plot = f"{dist_pix:.2f} pix"
            elif world_avail:
                c1 = SkyCoord(p1['world_ra'], p1['world_dec'], unit='deg', frame='icrs')
                c2 = SkyCoord(p2['world_ra'], p2['world_dec'], unit='deg', frame='icrs')
                text_plot = f"{c1.separation(c2).to(u.arcsec).value:.2f} arcsec"
            else:
                dist_pix = np.sqrt((p2.get('pixel_x', 0) - p1.get('pixel_x', 0))**2 +
                                   (p2.get('pixel_y', 0) - p1.get('pixel_y', 0))**2)
                text_plot = f"{dist_pix:.2f} pix"
            text_ui = text_plot
            if add_row:
                self._add_sky_distance_row(p1, p2, viewer)

        return text_ui, text_plot

    def _get_snap_coordinates(self, viewer):
        viewer_mark = self._get_mark(viewer)
        cursor_coords = self.coords_info.as_dict()
        cursor_x, cursor_y = cursor_coords.get('axes_x'), cursor_coords.get('axes_y')

        if not len(viewer_mark.x) or cursor_x is None or cursor_y is None:
            return cursor_coords

        marker_xs, marker_ys = viewer_mark.x, viewer_mark.y
        distances_sq = None

        if ('x' in viewer.scales and 'y' in viewer.scales and
                hasattr(viewer.scales['x'], 'domain') and hasattr(viewer.scales['x'], 'range') and
                hasattr(viewer.scales['y'], 'domain') and hasattr(viewer.scales['y'], 'range')):

            x_scale = viewer.scales['x']
            y_scale = viewer.scales['y']

            x_pixel_span = abs(x_scale.range[1] - x_scale.range[0])
            y_pixel_span = abs(y_scale.range[1] - y_scale.range[0])

            if x_pixel_span > 0 and y_pixel_span > 0:
                x_domain_span = x_scale.domain[1] - x_scale.domain[0]
                x_units_per_px = x_domain_span / x_pixel_span

                y_domain_span = y_scale.domain[1] - y_scale.domain[0]
                y_units_per_px = y_domain_span / y_pixel_span

                dx_pix = (marker_xs - cursor_x) / x_units_per_px
                dy_pix = (marker_ys - cursor_y) / y_units_per_px

                # Use 1D distance for spectrum viewers
                if not _is_image_viewer(viewer):
                    # For spectrum viewers, only consider the distance along the spectral axis.
                    distances_sq = dx_pix**2
                else:
                    # For image viewers, consider the 2d distance.
                    distances_sq = dx_pix**2 + dy_pix**2

        # If the preferred method was not possible or failed, use the fallback.
        if distances_sq is None:
            try:
                x_data_range = viewer.state.x_max - viewer.state.x_min
                y_data_range = viewer.state.y_max - viewer.state.y_min

                if not x_data_range or not y_data_range:
                    return cursor_coords

                norm_dx = (marker_xs - cursor_x) / x_data_range
                norm_dy = (marker_ys - cursor_y) / y_data_range

                if not _is_image_viewer(viewer):
                    distances_sq = norm_dx**2
                else:
                    distances_sq = norm_dx**2 + norm_dy**2

            except Exception:
                # If all else fails, do not snap
                return cursor_coords

        if distances_sq is None:
            # This case would be hit if both primary and fallback methods fail.
            return cursor_coords

        # Find the closest marker using the calculated distances.
        closest_marker_index_in_viewer = np.nanargmin(distances_sq)
        viewer_id = viewer.reference if viewer.reference is not None else viewer.reference_id
        viewer_loaded_data = [lyr.layer.label for lyr in viewer.layers]
        all_items_table = self.table.export_table()

        if all_items_table is None:
            return cursor_coords

        in_viewer_indices = [
            i for i, item in enumerate(all_items_table)
            if item.get('viewer') == viewer_id and item.get('data_label') in viewer_loaded_data
        ]

        if not in_viewer_indices or closest_marker_index_in_viewer >= len(in_viewer_indices):
            return cursor_coords

        snapped_table_index = in_viewer_indices[closest_marker_index_in_viewer]
        snapped_row = all_items_table[snapped_table_index]

        snapped_coords = {k: snapped_row.get(k) for k in self.table.headers_avail}
        snapped_coords['axes_x'] = viewer_mark.x[closest_marker_index_in_viewer]
        snapped_coords['axes_y'] = viewer_mark.y[closest_marker_index_in_viewer]
        return snapped_coords

    def _get_mark(self, viewer):
        matches = [mark for mark in viewer.figure.marks if isinstance(mark, MarkersMark)]
        if len(matches):
            return matches[0]
        mark = MarkersMark(viewer)
        viewer.figure.marks = viewer.figure.marks + [mark]
        return mark

    @property
    def marks(self):
        return {viewer_id: self._get_mark(viewer)
                for viewer_id, viewer in self.app._viewer_store.items()
                if hasattr(viewer, 'figure')}

    @property
    def coords_info(self):
        return self.app.session.application._tools['g-coords-info']

    @observe('is_active')
    def _on_is_active_changed(self, *args):
        if self.disabled_msg:
            return

        # toggle visibility of markers
        for mark in self.marks.values():
            mark.visible = self.is_active

        # subscribe/unsubscribe to keypress events across all viewers
        for viewer_marks in self._distance_marks.values():
            for mark in viewer_marks:
                mark.visible = self.is_active

        for viewer in self.app._viewer_store.values():
            if not hasattr(viewer, 'figure'):
                # table viewer, etc
                continue

            callback = self._viewer_callback(viewer, self._on_viewer_key_event)
            if self.is_active:
                viewer.add_event_callback(callback, events=['keydown'])
            else:
                viewer.remove_event_callback(callback)

    def _on_mouse_move_while_drawing(self, event):
        """
        Callback for 'mousemove' event. Updates the temporary DistanceMeasurement
        mark to follow the cursor.
        """
        if self._temporary_dist_measure is None or self._distance_first_point is None:
            return

        p1 = self._distance_first_point
        p2 = self.coords_info.as_dict()

        text_ui, text_plot = self._calculate_distance_text(p1, p2, self._viewer_for_drawing)
        self.distance_display = text_ui

        self._temporary_dist_measure.update_points(
            p1.get('axes_x'), p1.get('axes_y'),
            event['domain']['x'], event['domain']['y'],
            text=text_plot
        )

    # this is where items are being added to the table
    def _on_viewer_key_event(self, viewer, data):
        if data['event'] == 'keydown' and data['key'] == 'm':
            row_info = self.coords_info.as_dict()
            if 'viewer' in self.table.headers_avail:
                row_info['viewer'] = viewer.reference if viewer.reference is not None else viewer.reference_id # noqa

            for k in self.table.headers_avail:
                row_info.setdefault(k, self._default_table_values.get(k, ''))

            try:
                # if the pixel values are unreliable, set their table values as nan
                row_item_to_add = {k: float('nan') if row_info.get('pixel:unreliable', False) and
                                   k.startswith('pixel_') else v
                                   for k, v in row_info.items()
                                   if k in self.table.headers_avail}
                self.table.add_item(row_item_to_add)
                self.hub.broadcast(MarkersPluginUpdate(table_length=len(self.table), sender=self))

            except ValueError as err:  # pragma: no cover
                raise ValueError(f'failed to add {row_info} to table: {repr(err)}')
            x, y = row_info['axes_x'], row_info['axes_y']
            self._get_mark(viewer).append_xy(getattr(x, 'value', x), getattr(y, 'value', y))

        # on mac 'option + d' gets transmitted as delta instead of just d + alt
        elif data['event'] == 'keydown' and data.get('key') in ('d', '∂'):
            if data.get('altKey', False):
                coords = self._get_snap_coordinates(viewer)
            else:
                coords = self.coords_info.as_dict()
            viewer_id = viewer.reference or viewer.reference_id

            if self._distance_first_point is None:
                self._distance_first_point = coords
                self.distance_display = "..."
                self._viewer_for_drawing = viewer
                start_x, start_y = coords.get('axes_x'), coords.get('axes_y')

                is_profile = not _is_image_viewer(viewer)
                preview_text = "" if is_profile else "..."

                self._temporary_dist_measure = DistanceMeasurement(
                    viewer, x=[start_x, start_x], y=[start_y, start_y], text=preview_text
                )
                self._temporary_dist_measure.label_shadow.visible = False
                for mark in self._temporary_dist_measure.marks:
                    mark.opacities = [0.9]

                viewer.figure.marks = viewer.figure.marks + self._temporary_dist_measure.marks
                viewer.add_event_callback(self._on_mouse_move_while_drawing,
                                          events=['mousemove'])
            else:
                viewer.remove_event_callback(self._on_mouse_move_while_drawing)

                if self._temporary_dist_measure:
                    temp_marks = self._temporary_dist_measure.marks
                    viewer.figure.marks = [
                        m for m in viewer.figure.marks if m not in temp_marks
                    ]
                self._temporary_dist_measure = None

                p1 = self._distance_first_point
                p2 = coords
                self._distance_first_point = None

                text_ui, text_plot = self._calculate_distance_text(p1, p2, viewer, add_row=True)
                self.distance_display = text_ui

                plot_x0, plot_y0 = p1.get('axes_x'), p1.get('axes_y')
                plot_x1, plot_y1 = p2.get('axes_x'), p2.get('axes_y')

                if None in (plot_x0, plot_y0, plot_x1, plot_y1):
                    self.distance_display = "N/A"
                    self._viewer_for_drawing = None
                    return

                dist_measure = DistanceMeasurement(viewer, x=[plot_x0, plot_x1],
                                                   y=[plot_y0, plot_y1], text=text_plot)
                dist_measure.endpoints = {'p1': p1, 'p2': p2}
                self._distance_marks.setdefault(viewer_id, []).append(dist_measure)
                viewer.figure.marks = viewer.figure.marks + dist_measure.marks

                self._viewer_for_drawing = None



import os

from contextlib import contextmanager
from traitlets import Bool, Dict, Unicode, Integer, List, observe

from jdaviz.core.template_mixin import (SnackbarMessage, TemplateMixin, LayerSelect,
                                        LayerSelectMixin, DatasetSelectMixin)
from jdaviz.core.user_api import UserApiWrapper
from jdaviz.core.events import (AddDataMessage, ChangeRefDataMessage,
                                IconsUpdatedMessage, LayersFinalizedMessage,
                                LinkUpdatedMessage, ViewerRenamedMessage)
from glue.core.message import SubsetDeleteMessage
from jdaviz.core.sonified_layers import SonifiedLayerState, SonifiedDataLayerArtist
from jdaviz.utils import cmap_samples, is_not_wcs_only

from glue.core.edit_subset_mode import (AndMode, AndNotMode, OrMode,
                                        ReplaceMode, XorMode, NewMode)
from glue.core.roi import (RectangularROI, CircularROI,
                           EllipticalROI, CircularAnnulusROI, XRangeROI)
from glue.core.subset import CompositeSubsetState, RangeSubsetState, RoiSubsetState

from glue.icons import icon_path
from glue_jupyter.common.toolbar_vuetify import read_icon
from glue.viewers.scatter.state import ScatterLayerState
from glue_jupyter.bqplot.image import BqplotImageView

import ipyvuedraggable

__all__ = ['DataMenu']


SUBSET_TOOL_IDS = {'circle': 'bqplot:truecircle',
                   'rectangle': 'bqplot:rectangle',
                   'ellipse': 'bqplot:ellipse',
                   'annulus': 'bqplot:circannulus',
                   'xrange': 'bqplot:xrange',
                   'yrange': 'bqplot:yrange'}

SUBSET_NAMES = {v: k for k, v in SUBSET_TOOL_IDS.items()}

SUBSET_MODES = {
    'new': NewMode,
    'replace': ReplaceMode,
    'or': OrMode,
    'and': AndMode,
    'xor': XorMode,
    'andnot': AndNotMode,
}


class DataMenu(TemplateMixin, LayerSelectMixin, DatasetSelectMixin):
    """Viewer Data Menu

    Only the following attributes and methods are available through the
    :ref:`public API <plugin-apis>`:

    * :meth:`open_menu`
    * :meth:`data_labels_loaded`
    * :meth:`data_labels_visible`
    * :meth:`data_labels_unloaded`
    * ``layer`` (:class:`~jdaviz.core.template_mixin.LayerSelect`):
        actively selected layer(s)
    * ``orientation`` (:class:`~jdaviz.core.template_mixin.LayerSelect`):
        orientation chosen for the viewer, if applicable
    * :meth:`set_layer_visibility`
    * :meth:`toggle_layer_visibility`
    * :meth:`create_subset`
    * :meth:`modify_subset`
    * :meth:`resize_subset_in_viewer`
    * :meth:`add_data`
    * :meth:`view_info`
    * :meth:`remove_from_viewer`
    * :meth:`remove_from_app`
    """
    template_file = __file__, "data_menu.vue"

    force_open_menu = Bool(False).tag(sync=True)

    viewer_id = Unicode().tag(sync=True)
    viewer_reference = Unicode().tag(sync=True)

    icons = Dict().tag(sync=True)
    layer_icons = Dict().tag(sync=True)  # read-only, see app.state.layer_icons
    viewer_icons = Dict().tag(sync=True)  # read-only, see app.state.viewer_icons

    visible_layers = Dict().tag(sync=True)  # read-only, set by viewer

    orientation_enabled = Bool(False).tag(sync=True)
    orientation_align_by_wcs = Bool(False).tag(sync=True)
    orientation_layer_items = List().tag(sync=True)
    orientation_layer_selected = Unicode().tag(sync=True)

    viewer_supports_visible_toggle = Bool(True).tag(sync=True)
    disabled_layers_due_to_pixel_sky_mismatch = List().tag(sync=True)

    cmap_samples = Dict(cmap_samples).tag(sync=True)
    subset_tools = List().tag(sync=True)

    dm_layer_selected = List().tag(sync=True)

    loaded_n_data = Integer(0).tag(sync=True)
    selected_n_layers = Integer(0).tag(sync=True)
    selected_n_data = Integer(0).tag(sync=True)
    selected_n_subsets = Integer(0).tag(sync=True)

    info_enabled = Bool(False).tag(sync=True)
    info_tooltip = Unicode().tag(sync=True)

    delete_enabled = Bool(False).tag(sync=True)
    delete_tooltip = Unicode().tag(sync=True)
    delete_viewer_tooltip = Unicode().tag(sync=True)
    delete_app_enabled = Bool(False).tag(sync=True)
    delete_app_tooltip = Unicode().tag(sync=True)

    subset_edit_enabled = Bool(False).tag(sync=True)
    subset_edit_tooltip = Unicode().tag(sync=True)
    subset_resize_in_viewer_enabled = Bool(True).tag(sync=True)

    subset_edit_modes = List().tag(sync=True)

    def __init__(self, viewer, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # Trigger the loading of ipyvuedraggable in the frontend in notebook and Solara.
        ipyvuedraggable.Draggable()
        self._viewer = viewer
        self._during_select_sync = False

        from jdaviz.configs.default.plugins.viewers import JdavizTableViewer
        self.viewer_supports_visible_toggle = not isinstance(self._viewer, JdavizTableViewer)

        # TODO: refactor how this is applied by default to go through filters directly
        self.layer.remove_filter('filter_is_root')
        self.layer.add_filter(is_not_wcs_only)
        self.layer.multiselect = True
        self.layer.sort_by = 'zorder'
        self.layer._default_mode = 'empty'

        # we'll use a modified version of the dataset mixin to have a filtered
        # list of data entries in the app that are not in the current viewer.
        # changing the selection has no consequence.
        def data_not_in_viewer(data):
            return data.label not in self.layer.choices

        self.dataset.filters = ['is_not_wcs_only', 'not_child_layer',
                                data_not_in_viewer]

        self.orientation = LayerSelect(self,
                                       'orientation_layer_items',
                                       'orientation_layer_selected',
                                       'viewer_id',
                                       only_wcs_layers=True)

        self.orientation_enabled = (self.config in ('imviz', 'deconfigged')
                                    and viewer.__class__.__name__ == 'ImvizImageView')

        # first attach callback to catch any updates to viewer/layer icons and then
        # set their initial state
        self.hub.subscribe(self, IconsUpdatedMessage, self._on_app_icons_updated)
        self.hub.subscribe(self, AddDataMessage, handler=lambda _: self._set_viewer_id())
        self.hub.subscribe(self, LayersFinalizedMessage, handler=self._on_layers_finalized)
        self.hub.subscribe(self, LinkUpdatedMessage, handler=self._on_link_changed)
        self.hub.subscribe(self, SubsetDeleteMessage, handler=lambda msg: self._remove_subset_from_layers(msg.subset))  # noqa
        self.hub.subscribe(self, ChangeRefDataMessage, handler=self._on_refdata_change)
        self.hub.subscribe(self, ViewerRenamedMessage, handler=self._on_viewer_renamed_message)
        self.viewer_icons = dict(self.app.state.viewer_icons)
        self.layer_icons = dict(self.app.state.layer_icons)

        self.subset_edit_modes = [{'glue_name': 'replace', 'icon': read_icon(os.path.join(icon_path("glue_replace", icon_format="svg")), 'svg+xml')},  # noqa
                                  {'glue_name': 'or', 'icon': read_icon(os.path.join(icon_path("glue_or", icon_format="svg")), 'svg+xml')},  # noqa
                                  {'glue_name': 'and', 'icon': read_icon(os.path.join(icon_path("glue_and", icon_format="svg")), 'svg+xml')},  # noqa
                                  {'glue_name': 'xor', 'icon': read_icon(os.path.join(icon_path("glue_xor", icon_format="svg")), 'svg+xml')},  # noqa
                                  {'glue_name': 'andnot', 'icon': read_icon(os.path.join(icon_path("glue_andnot", icon_format="svg")), 'svg+xml')}]  # noqa

        # this currently assumes that toolbar.tools_data is set at init and does not change
        # if we ever support dynamic tool registration, this will need to be updated
        self.subset_tools = [{'id': k, 'img': v['img'], 'name': SUBSET_NAMES.get(k, k)}
                             for k, v in self._viewer.toolbar.tools_data.items()
                             if k in SUBSET_TOOL_IDS.values()]

        self.icons = {k: v for k, v in self.app.state.icons.items()}
        self.prevent_layer_items_recursion = False

    @property
    def user_api(self):
        expose = ['open_menu', 'layer', 'set_layer_visibility', 'toggle_layer_visibility',
                  'create_subset', 'modify_subset', 'resize_subset_in_viewer', 'add_data',
                  'view_info', 'remove_from_viewer', 'remove_from_app']
        if not self.viewer_supports_visible_toggle:
            expose = [e for e in expose
                      if e not in ('set_layer_visibility', 'toggle_layer_visibility',
                                   'create_subset', 'modify_subset')]
        readonly = ['data_labels_loaded', 'data_labels_visible', 'data_labels_unloaded']
        if self.app.config in ('imviz', 'deconfigged'):
            expose += ['orientation']
        return UserApiWrapper(self, expose=expose, readonly=readonly)

    def open_menu(self):
        """
        Open all instances of the data menu.
        """
        self.force_open_menu = True

    @property
    def existing_subset_labels(self):
        return [sg.label for sg in self.app.data_collection.subset_groups]

    @property
    def existing_sonified_labels(self):
        return [sd.layer.label for sd in self._viewer.state.layers
                if isinstance(sd, SonifiedLayerState)]

    @property
    def data_labels_loaded(self):
        return [layer['label'] for layer in self.layer_items
                if layer['label'] not in self.existing_subset_labels]

    @property
    def data_labels_visible(self):
        return [layer['label'] for layer in self.layer_items
                if layer['label'] not in self.existing_subset_labels and layer['visible']]

    @property
    def data_labels_unloaded(self):
        return self.dataset.choices

    def _remove_subset_from_layers(self, msg={}):
        if not hasattr(self, 'layer'):
            return

        # on SubsetDeleteMessage, message is sent before self.layer is updated, so update
        # layer visible to ensure icon is updated in data menu
        if msg.label not in self.existing_subset_labels:
            for layer in self._viewer.layers:
                if layer.layer.label == msg.label:
                    self.layer._update_items(remove_subset=msg.label)

    def _set_viewer_id(self):
        # viewer_ids are not populated on the viewer at init, so we'll keep checking and set
        # these the first time that they are available
        if len(self.viewer_id) and len(self.viewer_reference):
            return
        try:
            self.viewer_id = getattr(self._viewer, '_reference_id', '')
            self.viewer_reference = self._viewer.reference
            self.layer.viewer = self._viewer.reference
        except AttributeError:
            return
        self._on_refdata_change()

    def _on_app_icons_updated(self, msg):
        if msg.icon_type == 'viewer':
            self.viewer_icons = msg.icons
        elif msg.icon_type == 'layer':
            self.layer_icons = msg.icons
        self._set_viewer_id()

    def _on_layers_finalized(self, msg=None):
        # when a catalog is added, set visibility based on alignment type
        # and presence of pixel and/or world coordinates.
        new_data = msg.data
        if new_data.meta.get('_importer') == 'CatalogImporter':
            self._handle_catalog_layer_visibility(layers=[new_data])

    def _on_link_changed(self, msg=None):

        if len(self._viewer.layers):
            layers = [layer.layer for layer in self._viewer.layers]
            self._handle_catalog_layer_visibility(layers=layers)

    def _handle_catalog_layer_visibility(self, layers=None):
        """
        When a catalog layer is added, check if it should be visible based on
        presence of pixel/world coordinates and viewer alignment.

        If a catalog contains sky coordinates only and the viewer is aligned by
        pixels, hide the layer and show a snackbar message (and vice versa for
        world coordinates only and viewer aligned by WCS). If visibility is
        toggled off for this reason, the visibility toggle is greyed out in the
        data menu so it can not be toggled on again until the viewer alignment
        is changed.
        """

        if layers is None:
            return

        hiding_due_to_pixel_sky_mismatch = []
        coord_type = None
        align_type = None

        for layer in layers:
            if hasattr(layer, 'meta'):
                if layer.meta.get('_importer') == 'CatalogImporter':
                    ra_col = layer.meta.get('_jdaviz_loader_ra_col')
                    dec_col = layer.meta.get('_jdaviz_loader_dec_col')
                    x_col = layer.meta.get('_jdaviz_loader_x_col')
                    y_col = layer.meta.get('_jdaviz_loader_y_col')

                    comp_labels = [str(x) for x in layer.component_ids()]

                    has_world = ra_col in comp_labels and dec_col in comp_labels
                    has_pixel = x_col in comp_labels and y_col in comp_labels
                    align_by_wcs = self.orientation_align_by_wcs

                    if not ((align_by_wcs and has_world) or (not align_by_wcs and has_pixel)):  # noqa
                        coord_type = "pixel" if has_pixel else "world"
                        align_type = "WCS" if align_by_wcs else "Pixel"
                        self.set_layer_visibility(layer.label, False)
                        hiding_due_to_pixel_sky_mismatch.append(layer.label)

            if len(hiding_due_to_pixel_sky_mismatch):
                plural = 's' if len(hiding_due_to_pixel_sky_mismatch) > 1 else ''
                m = (f"Hiding Catalog layer{plural} {hiding_due_to_pixel_sky_mismatch} in "
                     f"'{self.viewer_reference}. Catalog{plural} contains {coord_type} "
                     f"coordinates only, which do not match current viewer alignment "
                     f"({align_type}).")
                self.hub.broadcast(SnackbarMessage(m, color='warning', sender=self))

        self.disabled_layers_due_to_pixel_sky_mismatch = hiding_due_to_pixel_sky_mismatch

    def _on_refdata_change(self, msg=None):
        if msg is not None and msg.viewer_id != self.viewer_id:
            return
        if getattr(self._viewer.state, 'reference_data', None) is None:
            return
        self.orientation_align_by_wcs = self._viewer.state.reference_data.meta.get('_WCS_ONLY', False)  # noqa
        if self.orientation_align_by_wcs:
            with self.during_select_sync():
                ref_label = self._viewer.state.reference_data.label
                if ref_label not in self.orientation.choices:
                    self.orientation._update_items()
                self.orientation.selected = str(ref_label)
        else:
            # update catalog layer visibility based on alignment / presence
            # of pixel/world coordinates
            if len(self._viewer.layers):
                layers = [layer.layer for layer in self._viewer.layers if layer.visible is True]
                self._handle_catalog_layer_visibility(layers)

    def _on_viewer_renamed_message(self, msg):
        if self.viewer_reference == msg.old_viewer_ref:
            self.viewer_reference = msg.new_viewer_ref
            self._set_viewer_id()

    @observe('orientation_layer_selected')
    def _orientation_layer_selected_changed(self, event={}):
        if not hasattr(self, 'orientation'):
            return
        if self._during_select_sync:
            return
        op = self._viewer.jdaviz_helper.plugins['Orientation']
        op._obj.set_orientation_for_viewer(event.get('new'), self.viewer_id)

    @contextmanager
    def during_select_sync(self):
        self._during_select_sync = True
        try:
            yield
        except Exception:  # pragma: no cover
            self._during_select_sync = False
            raise
        self._during_select_sync = False

    @observe('dm_layer_selected', 'layer_multiselect')
    def _dm_layer_selected_changed(self, event={}):
        if not hasattr(self, 'layer'):  # pragma: no cover
            return
        if self._during_select_sync:
            return
        if event.get('name') == 'layer_multiselect' and event.get('new'):
            return
        if not self.layer.multiselect and len(self.dm_layer_selected) > 1:
            # vue will still treat the element as a list, so we will include the
            # logic here to enforce single-select toggling
            self.dm_layer_selected = [self.dm_layer_selected[-1]]
            return
        if (event.get('name') == 'dm_layer_selected'
                and len(event.get('new')) == len(event.get('old'))):
            # not possible from UI interaction, but instead caused by a selected
            # layer being removed (deleting a selected subset, etc).  We want
            # to update dm_layer_selected in order to preserve layer.selected
            self._layers_changed(event)
            return
        with self.during_select_sync():
            # map index in dm_layer_selected (inverse order of layer_items)
            # to set self.layer.selected
            if len(self.layer_items):
                selected = [self.layer_items[i]['label']
                            for i in self.dm_layer_selected]
                if self.layer.multiselect:
                    self.layer.selected = selected
                else:
                    self.layer.selected = selected[0] if len(selected) else ''
            else:
                self.layer.selected = ''

    @observe('layer_selected', 'layer_items')
    def _layers_changed(self, event={}):
        # Avoid triggering this logic during app startup
        if not hasattr(self, '_viewer'):
            return

        # Have to count subsets + data, but not the invisible WCS layers in Imviz
        n_layers = len(set([lyr.layer.label for lyr in self._viewer.layers
                            if not (hasattr(lyr.layer, 'meta') and '_WCS_ONLY' in
                                    lyr.layer.meta and lyr.layer.meta['_WCS_ONLY'])]))

        if (event.get('name') == 'layer_items' and len(event['new']) == n_layers
                and isinstance(event.get('owner'), DataMenu)):
            # Setting layer.zorder causes a change to layer_items, which causes this function to be
            # immediately called again. Use this flag to prevent that.
            if self.prevent_layer_items_recursion:
                return

            self.prevent_layer_items_recursion = True

            label_order = [li['label'] for li in event["new"]
                           if li['is_subset'] is not None]
            not_in_order = [layer.layer.label for layer in self._viewer.layers
                            if layer.layer.label not in label_order]

            children_blocks = []
            data_and_children = []

            def _return_zorder_of_layers(layers):
                return [x.zorder for x in self._viewer.state.layers
                        if x.layer.label in layers]
            for data in self.app.data_collection:
                if self.app._get_assoc_data_children(data.label):
                    # If data has children, find their zorder blocks
                    # and return that as a list within a list of all
                    # data+children blocks
                    data_and_children.append(data.label)
                    data_and_children += (
                        self.app._get_assoc_data_children(data.label))
                    children_blocks.append(
                        _return_zorder_of_layers(
                            [data.label] +
                            self.app._get_assoc_data_children(data.label)))

            # Find old zorder of all layers previous to change
            old_layers = {item['label']: item['zorder'] for item in event['old']}

            # Build a map of child to parent for zorder adjustments
            child_to_parent = {child: data.label for data in self.app.data_collection
                               for child in self.app._get_assoc_data_children(data.label)}

            for layer in self._viewer.layers:
                # Skip reordering scatter layers in image viewers.
                if isinstance(self._viewer, BqplotImageView) and isinstance(layer.state, ScatterLayerState):  # noqa
                    continue
                if layer.layer.label in label_order:
                    new_zorder = len(label_order) - label_order.index(layer.layer.label)
                    if layer.layer.label not in data_and_children:
                        for x in children_blocks:
                            # If the projected new_zorder falls within a block of
                            # data+children, adjust to be just above or below the block
                            # depending on if the layer came from below or above, respectively
                            if min(x) <= new_zorder <= max(x):
                                # The old zorder was above the block and is moving down
                                if new_zorder < old_layers[layer.layer.label]:
                                    new_zorder = min(x)
                                # The old zorder was below the block and is moving up
                                else:
                                    new_zorder = max(x)
                    # If this is a child layer, ensure it stays above its parent
                    elif layer.layer.label in child_to_parent:
                        parent_label = child_to_parent[layer.layer.label]
                        # Find the parent's zorder
                        parent_zorder = None
                        for parent_layer in self._viewer.layers:
                            if parent_layer.layer.label == parent_label:
                                parent_zorder = parent_layer.zorder
                                break
                        # Child must have higher zorder than parent (rendered on top)
                        if parent_zorder is not None and new_zorder <= parent_zorder:
                            new_zorder = parent_zorder + 0.1
                else:
                    new_zorder = len(not_in_order) + len(label_order) - not_in_order.index(layer.layer.label)  # noqa

                if new_zorder != layer.zorder:
                    layer.zorder = new_zorder
                    self.layer._update_items()

            self.prevent_layer_items_recursion = False

            # Only trigger if the order expected in the message and the actual order differ
            if label_order != [li['label'] for li in self.layer_items if li['is_subset'] is not None]:  # noqa
                self.layer._update_items()

        if not hasattr(self, 'layer') or not self.layer.multiselect:  # pragma: no cover
            return
        if not self._during_select_sync:
            with self.during_select_sync():
                # map list of strings in self.layer.selected to indices in dm_layer_selected
                layer_labels = [layer['label'] for layer in self.layer_items]
                layer_selected = self.layer_selected if self.layer.multiselect else [self.layer_selected]  # noqa
                self.dm_layer_selected = [layer_labels.index(label) for label in layer_selected
                                          if label in layer_labels]

        subset_labels = self.existing_subset_labels

        if event.get('name') == 'layer_items':
            # changing the layers in the viewer needs to trigger an update to dataset_items
            # through the set filters
            self.dataset._update_items()
            self.loaded_n_data = len([lyr for lyr in self.layer.choices
                                      if lyr not in subset_labels])
            return

        # update internal counts and tooltips
        self.selected_n_layers = len(self.layer.selected)
        self.selected_n_subsets = len([lyr for lyr in self.layer.selected if lyr in subset_labels])
        self.selected_n_data = self.selected_n_layers - self.selected_n_subsets

        # user-friendly representation of selection
        selected_repr = ""
        if self.selected_n_data:
            selected_repr += f"data ({self.selected_n_data})"
        if self.selected_n_subsets:
            if self.selected_n_data:
                selected_repr += " and"
            if self.selected_n_subsets == 1:
                selected_repr += f" subset ({self.selected_n_subsets})"
            else:
                selected_repr += f" subsets ({self.selected_n_subsets})"

        # layer info rules
        if self.selected_n_layers == 1:
            if self.dm_layer_selected == [] or max(self.dm_layer_selected) >= len(self.layer_items):  # noqa pragma: no cover
                # Can happen during state transition but should immediately be followed up
                # with an update. Also get here when unloading all data.
                self.info_enabled = False
                self.info_tooltip = ''
            elif self.layer_items[self.dm_layer_selected[0]].get('from_plugin', False):
                self.info_enabled = False
                self.info_tooltip = 'Selected data layer is a plugin product and does not have metadata'  # noqa
            else:
                self.info_enabled = True
                if self.selected_n_data == 1:
                    self.info_tooltip = 'View metadata for selected data'
                else:
                    self.info_tooltip = 'View subset info for selected subset'
        else:
            self.info_enabled = False
            if self.selected_n_layers == 0:
                self.info_tooltip = 'Select a layer to view info'
            else:
                self.info_tooltip = 'Select a single layer to view info'

        # delete layer rules
        if self.selected_n_layers == 0:
            self.delete_tooltip = "Select layer(s) to delete"
            self.delete_enabled = False
        else:
            self.delete_tooltip = f"Remove selected {selected_repr}"
            self.delete_enabled = True

        # delete from entire app rules
        subset_str = "subset" if self.selected_n_subsets == 1 else "subsets"
        if self.selected_n_subsets and self.selected_n_data:
            self.delete_viewer_tooltip = f"Remove selected data and hide selected {subset_str} in this viewer"  # noqa
        elif self.selected_n_data:
            self.delete_viewer_tooltip = "Remove selected data from this viewer"
        elif self.selected_n_subsets:
            self.delete_viewer_tooltip = f"Hide selected {subset_str} in this viewer"

        delete_app_tooltip = "Remove from all viewers and application (permanent, might affect existing subsets)"  # noqa
        if self.app.config == 'cubeviz':
            # forbid deleting non-plugin generated data
            selected_items = self.layer.selected_item
            for i, layer in enumerate(self.layer.selected):
                if layer == '':
                    # In this case we removed everything from the viewer
                    continue
                if isinstance(self.layer.selected_obj[0][0], SonifiedDataLayerArtist):
                    self.delete_app_enabled = False
                    self.delete_app_tooltip = "Cannot delete sonified data from app"
                    break
                elif (layer not in self.existing_subset_labels
                      and selected_items['from_plugin'][i] is None):
                    self.delete_app_enabled = False
                    self.delete_app_tooltip = f"Cannot delete imported data from {self.app.config}"
                    break
            else:
                self.delete_app_enabled = True
                self.delete_app_tooltip = delete_app_tooltip
        else:
            self.delete_app_enabled = True
            self.delete_app_tooltip = delete_app_tooltip

        # subset edit rules
        if self.selected_n_subsets == 1 and self.selected_n_layers == 1:
            self.subset_edit_enabled = True
            self.subset_edit_tooltip = f"Edit {self.layer_selected[0]}"
        else:
            self.subset_edit_enabled = False
            if self.selected_n_subsets == 0:
                self.subset_edit_tooltip = "Select a subset to edit"
            else:
                self.subset_edit_tooltip = "Select a single subset to edit"
        self.layer._update_items()

    def set_layer_visibility(self, layer_label, visible=True):
        """
        Set the visibility of a layer in the viewer.

        Parameters
        ----------
        layer_label : str
            The label of the layer to set the visibility of.
        visible : bool
            Whether the layer should be visible or not.

        Returns
        -------
        dict
            A dictionary of the current visible layers.
        """
        if visible and layer_label in self.disabled_layers_due_to_pixel_sky_mismatch:
            raise ValueError(f"Layer '{layer_label}' cannot be made visible when "
                             "viewer is aligned by pixel coordinates.")
        for layer in self._viewer.layers:
            if layer.layer.label == layer_label:
                if isinstance(layer, SonifiedDataLayerArtist):
                    layer.audible = visible
                else:
                    layer.visible = visible
            elif hasattr(layer.layer, 'data') and layer.layer.data.label == layer_label:
                layer.visible = layer.layer.label in self.visible_layers
            if not visible and self.app._get_assoc_data_parent(layer.layer.label) == layer_label:
                # then this is a child-layer of a parent-layer that is being hidden
                # so also hide the child-layer
                layer.visible = False

            if layer.state.visible != layer.visible:
                layer.state.visible = layer.visible

        # Update items once after all visibility changes are made
        self.layer._update_items()

        if visible and (parent_label := self.app._get_assoc_data_parent(layer_label)):
            # ensure the parent layer is also visible
            self.set_layer_visibility(parent_label, visible=True)

        return self.visible_layers

    def toggle_layer_visibility(self, layer_label):
        """
        Toggle the visibility of a layer in the viewer.

        Parameters
        ----------
        layer_label : str
            The label of the layer to toggle the visibility of.

        Returns
        -------
        bool
            The new visibility state of the layer.
        """
        visible = layer_label not in self.visible_layers
        self.set_layer_visibility(layer_label, visible=visible)
        return visible

    def vue_set_layer_visibility(self, info, *args):
        return self.set_layer_visibility(info.get('layer'), info.get('value'))  # pragma: no cover

    def add_data(self, *data_labels):
        """
        Add a dataset to the viewer.

        Parameters
        ----------
        *data_labels : str
            The label(s) of the dataset to add to the viewer.
        """
        unavailable = [data_label for data_label in data_labels
                       if data_label not in self.dataset.choices]
        if len(unavailable):
            raise ValueError(f"Data labels {unavailable} not able to be loaded into '{self.viewer_id}'.  Must be one of: {self.dataset.choices}")  # noqa
        for data_label in data_labels:
            self.app.add_data_to_viewer(self.viewer_id, data_label)

    def vue_add_data_to_viewer(self, info, *args):
        self.add_data(info.get('data_label'))  # pragma: no cover

    def create_subset(self, subset_type):
        """
        Interactively create a new subset in the viewer.  This sets the app-wide subset
        selection to 'Create New' and selects the appropriate tool in this viewer's toolbar.

        Parameters
        ----------
        subset_type : str
            The type of subset to create.  Must be one of 'circle', 'rectangle', 'ellipse',
            'annulus', 'xrange', or 'yrange', and must be an available tool in this viewer.
        """
        # clear previous selection, finalize subsets, temporarily sets default tool
        self._viewer.toolbar.active_tool_id = None
        # set toolbar to the selection, will also set app-wide subset selection to "Create New"
        # NOTE: supports passing either the user-friendly name or the actual ID
        self._viewer.toolbar.select_tool(SUBSET_TOOL_IDS.get(subset_type, subset_type))

    def vue_create_subset(self, info, *args):
        self.create_subset(info.get('subset_type'))  # pragma: no cover

    def modify_subset(self, combination_mode, subset_type):
        """
        Interactively modify an existing subset in the viewer.  This sets the app-wide subset
        selection to the currently selected subset, mode to the selected combination_mode,
        and selects the appropriate tool in this viewer's toolbar.

        Parameters
        ----------
        combination_mode : str
            The combination mode to apply to the subset.  Must be one of 'replace', 'or', 'and',
            'xor', or 'andnot'.
        subset_type : str
            The type of subset to modify.  Must be one of 'circle', 'rectangle', 'ellipse',
            'annulus', 'xrange', or 'yrange', and must be an available tool in this viewer.
        """
        # future improvement: allow overriding layer.selected, with pre-validation
        if len(self.layer.selected) != 1:
            raise ValueError("Only one layer can be selected to modify subset.")
        if self.layer.selected[0] not in self.existing_subset_labels:
            raise ValueError("Selected layer is not a subset.")
        subset = self.layer.selected[0]

        # set tool first since that might default to "Create New"
        self._viewer.toolbar.select_tool(SUBSET_TOOL_IDS.get(subset_type, subset_type))
        # set subset selection to the subset to modify
        subset_grp = [sg for sg in self.app.data_collection.subset_groups if sg.label == subset]
        self.session.edit_subset_mode.edit_subset = subset_grp
        # set combination mode
        self.session.edit_subset_mode.mode = SUBSET_MODES.get(combination_mode)

    def vue_modify_subset(self, info, *args):
        self.modify_subset(info.get('combination_mode'),
                           info.get('subset_type'))  # pragma: no cover

    @observe('layer_selected')
    def _allow_resize_subset_in_viewer(self, event={}):
        # Needed during app initialization
        if not hasattr(self, 'layer'):
            return
        if not len(self.layer.selected):
            return

        subset_grp = [sg for sg in self.app.data_collection.subset_groups
                      if sg.label == self.layer.selected[0]]
        if len(subset_grp):
            subset = subset_grp[0]
            if isinstance(subset.subset_state, CompositeSubsetState):
                self.subset_resize_in_viewer_enabled = False
            else:
                self.subset_resize_in_viewer_enabled = True

    def resize_subset_in_viewer(self):
        """
        Enable resizing of an existing subset in the viewer.
        Detects the subset's ROI type and activates the corresponding
        selection tool pre-populated with the existing bounds for
        interactive resizing.
        """
        # future improvement: allow overriding layer.selected, with pre-validation
        if len(self.layer.selected) != 1:
            raise ValueError('Only one layer can be selected to resize subset.')
        if self.layer.selected[0] not in self.existing_subset_labels:
            raise ValueError('Selected layer is not a subset.')
        if not self.subset_resize_in_viewer_enabled:
            raise ValueError('Resizing is not supported for composite subsets.')

        subset_str = self.layer.selected[0]

        # set subset as the active/highlighted layer in data menu
        self.layer.selected = subset_str

        # get the subset group and extract the ROI
        subset_grp = [sg for sg in self.app.data_collection.subset_groups
                      if sg.label == subset_str]

        # Determine which tool to use based on the ROI type
        if len(subset_grp):
            subset = subset_grp[0]
            if isinstance(subset.subset_state, RoiSubsetState):
                roi = subset.subset_state.roi
            elif isinstance(subset.subset_state, RangeSubsetState):
                roi = XRangeROI(subset.subset_state.lo, subset.subset_state.hi)
            else:
                raise ValueError('Selected subset does not have a supported ROI.')

            # Map ROI types to tool IDs
            roi_tool_map = {RectangularROI: 'bqplot:rectangle',
                            CircularROI: 'bqplot:truecircle',
                            EllipticalROI: 'bqplot:ellipse',
                            CircularAnnulusROI: 'bqplot:circannulus',
                            XRangeROI: 'bqplot:xrange'}

            # Find the matching tool for this ROI type
            tool_id = None
            for roi_type, tid in roi_tool_map.items():
                if isinstance(roi, roi_type):
                    tool_id = tid
                    break

            if tool_id is None:
                raise NotImplementedError(
                    f'Resize not supported for {roi.__class__.__name__} subsets.')

            # Activate the appropriate tool
            self._viewer.toolbar.active_tool_id = None
            self._viewer.toolbar.select_tool(tool_id)

            # Set subset to edit and mode to replace
            self.session.edit_subset_mode.edit_subset = subset_grp
            self.session.edit_subset_mode.mode = ReplaceMode

            # Pre-populate the tool with the existing subset's ROI
            tool = self._viewer.toolbar.active_tool
            if tool and hasattr(tool, 'update_from_roi'):
                tool.update_from_roi(roi)
            else:
                raise RuntimeError(
                    f'Failed to activate {tool.__class__.__name__} '
                    f'for resizing {roi.__class__.__name__} subsets.')

    def vue_resize_subset_in_viewer(self, *args):
        self.resize_subset_in_viewer()  # pragma: no cover

    def view_info(self):
        """
        View info for the selected layer by opening either the metadata or subset plugin to the
        selected entry.
        """
        # future improvement: allow overriding layer.selected, with pre-validation
        if len(self.layer.selected) != 1:
            raise ValueError("Only one layer can be selected to view info.")
        if self.layer.selected[0] in self.existing_subset_labels:
            sp = self._viewer.jdaviz_helper.plugins.get('Subset Tools', None)
            if sp is None:  # pragma: no cover
                raise ValueError("subset tools plugin not available")
            sp._obj.subset.selected = self.layer.selected[0]
            sp.open_in_tray()
        else:
            mp = self._viewer.jdaviz_helper.plugins.get('Metadata', None)
            if mp is None:  # pragma: no cover
                raise ValueError("metadata plugin not available")
            mp.dataset.selected = self.layer.selected[0]
            mp.open_in_tray()

    def vue_view_info(self, *args):
        self.view_info()  # pragma: no cover

    def remove_from_viewer(self):
        """
        Remove the selected layers from the viewer.  For subset layers, this
        sets the visibility of the subset layer.  For data layers,
        this unloads the data from the viewer, but keeps it in the application or other viewers.
        """
        # future improvement: allow overriding layer.selected via *args, with pre-validation
        for layer in self.layer.selected:
            if layer in self.existing_subset_labels:
                self.set_layer_visibility(layer, visible=False)
            else:
                self.app.remove_data_from_viewer(self.viewer_id, layer)

    def vue_remove_from_viewer(self, *args):
        self.remove_from_viewer()  # pragma: no cover

    def remove_from_app(self):
        """
        Remove the selected layers from the entire app and all viewers.
        """
        # future improvement: allow overriding layer.selected via *args, with pre-validation
        for layer in (self.layer.selected if isinstance(self.layer.selected,
                                                        (list, tuple)) else [self.layer.selected]):
            if layer in self.existing_subset_labels:
                for sg in self.app.data_collection.subset_groups:
                    if sg.label == layer:
                        self.app.data_collection.remove_subset_group(sg)
                        self.layer._update_items()
                        break
            else:
                self.app.data_item_remove(layer)

    def vue_remove_from_app(self, *args):
        self.remove_from_app()  # pragma: no cover

    def vue_open_orientation_plugin(self, *args):
        op = self._viewer.jdaviz_helper.plugins.get('Orientation', None)
        if op is None:  # pragma: no cover
            raise ValueError("orientation plugin not available")
        op.open_in_tray()




import pytest
from jdaviz.utils import cached_uri


def test_load_rectangular_ramp(rampviz_helper, jwst_level_1b_rectangular_ramp):
    rampviz_helper.load_data(jwst_level_1b_rectangular_ramp)

    # drop the integration axis
    original_cube_shape = jwst_level_1b_rectangular_ramp.shape[1:]

    # on ramp cube load (1), the parser loads a diff cube (2) and
    # the ramp extraction plugin produces a default extraction (3):
    assert len(rampviz_helper.app.data_collection) == 3

    parsed_cube_shape = rampviz_helper.app.data_collection[0].shape
    assert parsed_cube_shape == (
        original_cube_shape[1], original_cube_shape[2], original_cube_shape[0]
    )


def test_load_level_1_and_2(
        rampviz_helper,
        jwst_level_1b_rectangular_ramp,
        jwst_level_2c_rate_image
):
    # load level 1 ramp and level 2 rate image
    rampviz_helper.load_data(jwst_level_1b_rectangular_ramp)
    rampviz_helper.load_data(jwst_level_2c_rate_image)

    # confirm that a "level-2" viewer is created, and that
    # the rate image is loaded into it
    assert len(rampviz_helper.viewers) == 4
    assert 'level-2' in rampviz_helper.viewers

    layers = rampviz_helper.app.get_viewer('level-2').layers
    assert len(layers) == 1


@pytest.mark.remote_data
@pytest.mark.parametrize("url", [
    # "https://stsci.box.com/shared/static/vklnig8f7fflyfwrfa9t6vpl5vnoi5mb.asdf",
    "mast:JWST/product/jw01181003001_08201_00003_mirimage_uncal.fits",
    "mast:JWST/product/jw03383196001_04201_00004_nis_uncal.fits"
])
def test_load_example_notebook_data(rampviz_helper, url):
    try:
        import roman_datamodels  # noqa
    except ImportError:
        has_rdd = False
    else:
        has_rdd = True
    uri = cached_uri(url)
    if 'mast' in uri:
        ldr = rampviz_helper.loaders['url']
        ldr.cache = True
        ldr.url = uri
    else:
        ldr = rampviz_helper.loaders['file']
        ldr.filepath = uri

    if url.endswith(".asdf") and not has_rdd:
        assert 'Ramp' not in ldr.format.choices
    else:
        assert 'Ramp' in ldr.format.choices
        ldr.format = 'Ramp'
        ldr.load()



"""The ``region_translators`` module houses translations of
:ref:`regions:shapes` to :ref:`photutils:photutils-aperture` apertures.
"""
import re
from astropy import units as u
from astropy.coordinates import SkyCoord
from glue.core.roi import CircularROI, EllipticalROI, RectangularROI, CircularAnnulusROI
from photutils.aperture import (CircularAperture, SkyCircularAperture,
                                EllipticalAperture, SkyEllipticalAperture,
                                RectangularAperture, SkyRectangularAperture,
                                CircularAnnulus, SkyCircularAnnulus,
                                EllipticalAnnulus, SkyEllipticalAnnulus,
                                RectangularAnnulus, SkyRectangularAnnulus)
from regions import (CirclePixelRegion, CircleSkyRegion,
                     EllipsePixelRegion, EllipseSkyRegion,
                     RectanglePixelRegion, RectangleSkyRegion,
                     CircleAnnulusPixelRegion, CircleAnnulusSkyRegion,
                     EllipseAnnulusPixelRegion, EllipseAnnulusSkyRegion,
                     RectangleAnnulusPixelRegion, RectangleAnnulusSkyRegion,
                     PixCoord, PolygonSkyRegion)

__all__ = ['regions2roi', 'regions2aperture', 'aperture2regions']


def regions2roi(region_shape, wcs=None):
    """Convert a given ``regions`` shape to ``glue`` ROI.

    This is the opposite of what is offered by
    ``glue_astronomy.translators.regions.AstropyRegionsHandler.to_object``
    but does not cover all the same shapes exactly.

    Parameters
    ----------
    region_shape : `regions.Region`
        A supported ``regions`` shape.

    wcs : `~astropy.wcs.WCS` or `None`
        A compatible WCS object, if required.
        **This is only used for sky aperture.**

    Returns
    -------
    roi : `glue.core.roi.Roi`
        An equivalent ``glue`` ROI.

    Raises
    ------
    ValueError
        WCS is required but not provided.

    NotImplementedError
        The given ``regions`` shape is not supported.

    Examples
    --------
    Translate a `regions.CirclePixelRegion` to `glue.core.roi.CircularROI`:

    >>> from regions import CirclePixelRegion, PixCoord
    >>> from jdaviz.core.region_translators import regions2roi
    >>> region_shape = CirclePixelRegion(center=PixCoord(x=42, y=43), radius=4.2)
    >>> regions2roi(region_shape)  # doctest: +ELLIPSIS
    <glue.core.roi.CircularROI object at ...>

    """
    if isinstance(region_shape, (CircleSkyRegion, EllipseSkyRegion, RectangleSkyRegion,
                                 CircleAnnulusSkyRegion)):
        if wcs is None:
            raise ValueError(f'WCS must be provided for {region_shape}')

        # Convert sky to pixel region first, if necessary.
        region_shape = region_shape.to_pixel(wcs)

    if isinstance(region_shape, CirclePixelRegion):
        roi = CircularROI(
            xc=region_shape.center.x, yc=region_shape.center.y, radius=region_shape.radius)
    elif isinstance(region_shape, EllipsePixelRegion):
        roi = EllipticalROI(
            xc=region_shape.center.x, yc=region_shape.center.y,
            radius_x=region_shape.width * 0.5, radius_y=region_shape.height * 0.5,
            theta=region_shape.angle.to_value(u.radian))
    elif isinstance(region_shape, RectanglePixelRegion):
        half_w = region_shape.width * 0.5
        half_h = region_shape.height * 0.5
        xmin = region_shape.center.x - half_w
        xmax = region_shape.center.x + half_w
        ymin = region_shape.center.y - half_h
        ymax = region_shape.center.y + half_h
        roi = RectangularROI(
            xmin=xmin, xmax=xmax, ymin=ymin, ymax=ymax,
            theta=region_shape.angle.to_value(u.radian))
    elif isinstance(region_shape, CircleAnnulusPixelRegion):
        roi = CircularAnnulusROI(
            xc=region_shape.center.x, yc=region_shape.center.y,
            inner_radius=region_shape.inner_radius, outer_radius=region_shape.outer_radius)
    else:
        raise NotImplementedError(f'{region_shape.__class__.__name__} is not supported')

    return roi


def regions2aperture(region_shape):
    """Convert a given ``regions`` shape to ``photutils`` aperture.

    Parameters
    ----------
    region_shape : `regions.Region`
        A supported ``regions`` shape.

    Returns
    -------
    aperture : `photutils.aperture.Aperture`
        An equivalent ``photutils`` aperture.

    Raises
    ------
    NotImplementedError
        The given ``regions`` shape is not supported.

    Examples
    --------
    Translate a `regions.CirclePixelRegion` to `photutils.aperture.CircularAperture`:

    >>> from regions import CirclePixelRegion, PixCoord
    >>> from jdaviz.core.region_translators import regions2aperture
    >>> region_shape = CirclePixelRegion(center=PixCoord(x=42, y=43), radius=4.2)
    >>> regions2aperture(region_shape)
    <CircularAperture([42., 43.], r=4.2)>

    See Also
    --------
    aperture2regions

    """
    if isinstance(region_shape, CirclePixelRegion):
        aperture = CircularAperture(region_shape.center.xy, region_shape.radius)

    elif isinstance(region_shape, CircleSkyRegion):
        aperture = SkyCircularAperture(region_shape.center, region_shape.radius)

    elif isinstance(region_shape, EllipsePixelRegion):
        th = region_shape.angle
        aperture = EllipticalAperture(
            region_shape.center.xy, region_shape.width * 0.5, region_shape.height * 0.5,
            theta=th)

    elif isinstance(region_shape, EllipseSkyRegion):
        aperture = SkyEllipticalAperture(
            region_shape.center, region_shape.width * 0.5, region_shape.height * 0.5,
            theta=(region_shape.angle - (90 * u.deg)))

    elif isinstance(region_shape, RectanglePixelRegion):
        th = region_shape.angle
        aperture = RectangularAperture(
            region_shape.center.xy, region_shape.width, region_shape.height,
            theta=th)

    elif isinstance(region_shape, RectangleSkyRegion):
        aperture = SkyRectangularAperture(
            region_shape.center, region_shape.width, region_shape.height,
            theta=(region_shape.angle - (90 * u.deg)))

    elif isinstance(region_shape, CircleAnnulusPixelRegion):
        aperture = CircularAnnulus(
            region_shape.center.xy, region_shape.inner_radius, region_shape.outer_radius)

    elif isinstance(region_shape, CircleAnnulusSkyRegion):
        aperture = SkyCircularAnnulus(
            region_shape.center, region_shape.inner_radius, region_shape.outer_radius)

    elif isinstance(region_shape, EllipseAnnulusPixelRegion):
        th = region_shape.angle
        aperture = EllipticalAnnulus(
            region_shape.center.xy, region_shape.inner_width * 0.5, region_shape.outer_width * 0.5,
            region_shape.outer_height * 0.5, b_in=region_shape.inner_height * 0.5,
            theta=th)

    elif isinstance(region_shape, EllipseAnnulusSkyRegion):
        aperture = SkyEllipticalAnnulus(
            region_shape.center, region_shape.inner_width * 0.5, region_shape.outer_width * 0.5,
            region_shape.outer_height * 0.5, b_in=region_shape.inner_height * 0.5,
            theta=(region_shape.angle - (90 * u.deg)))

    elif isinstance(region_shape, RectangleAnnulusPixelRegion):
        th = region_shape.angle
        aperture = RectangularAnnulus(
            region_shape.center.xy, region_shape.inner_width, region_shape.outer_width,
            region_shape.outer_height, h_in=region_shape.inner_height,
            theta=th)

    elif isinstance(region_shape, RectangleAnnulusSkyRegion):
        aperture = SkyRectangularAnnulus(
            region_shape.center, region_shape.inner_width, region_shape.outer_width,
            region_shape.outer_height, h_in=region_shape.inner_height,
            theta=(region_shape.angle - (90 * u.deg)))

    else:
        raise NotImplementedError(f'{region_shape.__class__.__name__} is not supported')

    return aperture


def aperture2regions(aperture):
    """Convert a given ``photutils`` aperture to ``regions`` shape.

    Parameters
    ----------
    aperture : `photutils.aperture.Aperture`
        An equivalent ``photutils`` aperture.

    Returns
    -------
    region_shape : `regions.Region`
        A supported ``regions`` shape.

    Raises
    ------
    NotImplementedError
        The given ``photutils`` aperture is not supported.

    ValueError
        Invalid inputs.

    Examples
    --------
    Translate a `photutils.aperture.CircularAperture` to `regions.CirclePixelRegion`:

    >>> from photutils.aperture import CircularAperture
    >>> from jdaviz.core.region_translators import aperture2regions
    >>> aperture = CircularAperture((42, 43), 4.2)
    >>> aperture2regions(aperture)
    <CirclePixelRegion(center=PixCoord(x=42.0, y=43.0), radius=4.2)>

    See Also
    --------
    regions2aperture

    """
    if isinstance(aperture, CircularAperture):
        region_shape = CirclePixelRegion(
            center=positions2pixcoord(aperture.positions), radius=aperture.r)

    elif isinstance(aperture, SkyCircularAperture):
        region_shape = CircleSkyRegion(center=aperture.positions, radius=aperture.r)

    elif isinstance(aperture, EllipticalAperture):
        region_shape = EllipsePixelRegion(
            center=positions2pixcoord(aperture.positions), width=aperture.a * 2,
            height=aperture.b * 2, angle=aperture.theta)

    elif isinstance(aperture, SkyEllipticalAperture):
        region_shape = EllipseSkyRegion(
            center=aperture.positions, width=aperture.a * 2, height=aperture.b * 2,
            angle=(aperture.theta + (90 * u.deg)))

    elif isinstance(aperture, RectangularAperture):
        region_shape = RectanglePixelRegion(
            center=positions2pixcoord(aperture.positions), width=aperture.w, height=aperture.h,
            angle=aperture.theta)

    elif isinstance(aperture, SkyRectangularAperture):
        region_shape = RectangleSkyRegion(
            center=aperture.positions, width=aperture.w, height=aperture.h,
            angle=(aperture.theta + (90 * u.deg)))

    elif isinstance(aperture, CircularAnnulus):
        region_shape = CircleAnnulusPixelRegion(
            center=positions2pixcoord(aperture.positions), inner_radius=aperture.r_in,
            outer_radius=aperture.r_out)

    elif isinstance(aperture, SkyCircularAnnulus):
        region_shape = CircleAnnulusSkyRegion(
            center=aperture.positions, inner_radius=aperture.r_in, outer_radius=aperture.r_out)

    elif isinstance(aperture, EllipticalAnnulus):
        region_shape = EllipseAnnulusPixelRegion(
            center=positions2pixcoord(aperture.positions), inner_width=aperture.a_in * 2,
            inner_height=aperture.b_in * 2, outer_width=aperture.a_out * 2,
            outer_height=aperture.b_out * 2, angle=aperture.theta)

    elif isinstance(aperture, SkyEllipticalAnnulus):
        region_shape = EllipseAnnulusSkyRegion(
            center=aperture.positions, inner_width=aperture.a_in * 2,
            inner_height=aperture.b_in * 2, outer_width=aperture.a_out * 2,
            outer_height=aperture.b_out * 2,
            angle=(aperture.theta + (90 * u.deg)))

    elif isinstance(aperture, RectangularAnnulus):
        region_shape = RectangleAnnulusPixelRegion(
            center=positions2pixcoord(aperture.positions), inner_width=aperture.w_in,
            inner_height=aperture.h_in, outer_width=aperture.w_out, outer_height=aperture.h_out,
            angle=aperture.theta)

    elif isinstance(aperture, SkyRectangularAnnulus):
        region_shape = RectangleAnnulusSkyRegion(
            center=aperture.positions, inner_width=aperture.w_in, inner_height=aperture.h_in,
            outer_width=aperture.w_out, outer_height=aperture.h_out,
            angle=(aperture.theta + (90 * u.deg)))

    else:  # pragma: no cover
        raise NotImplementedError(f'{aperture.__class__.__name__} is not supported')

    return region_shape


def _create_polygon_skyregion_from_coords(coords, frame='icrs', unit=u.deg):
    """Create a `regions.PolygonSkyRegion` from given coordinates.

    Parameters
    ----------
    coords : list
        List of coordinates in the form of [ra1, dec1, ra2, dec2, ...].

    frame : str, optional
        Coordinate frame. Default is 'icrs'.

    unit : `~astropy.units.Unit`, optional
        Unit of the coordinates. Default is `~astropy.units.deg`.

    Returns
    -------
    sky_region : `regions.PolygonSkyRegion`
        A polygon sky region.

    Examples
    --------
    Create a `regions.PolygonSkyRegion` from given coordinates:

    >>> from jdaviz.core.region_translators import _create_polygon_skyregion_from_coords
    >>> coords = [10.0, 20.0, 30.0, 40.0, 50.0, 60.0]
    >>> _create_polygon_skyregion_from_coords(coords)
    <PolygonSkyRegion(vertices=<SkyCoord (ICRS): (ra, dec) in deg
        [(10., 20.), (30., 40.), (50., 60.)]>)>

    """
    ra, dec = coords[::2], coords[1::2]
    sky_coordinates = SkyCoord(ra, dec, frame=frame, unit=unit)
    sky_region = PolygonSkyRegion(sky_coordinates)

    return sky_region


def _create_circle_skyregion_from_coords(coords, frame='icrs', unit=u.deg):
    """Create a `regions.CircleSkyRegion` from given coordinates.

    Parameters
    ----------
    coords : list
        List of coordinates in the form of [ra, dec, radius].

    frame : str, optional
        Coordinate frame. Default is 'icrs'.

    unit : `~astropy.units.Unit`, optional
        Unit of the coordinates. Default is `~astropy.units.deg`.

    Returns
    -------
    sky_region : `regions.CircleSkyRegion`
        A circle sky region.

    Examples
    --------
    Create a `regions.CircleSkyRegion` from given coordinates:

    >>> from jdaviz.core.region_translators import _create_circle_skyregion_from_coords
    >>> coords = [10.0, 20.0, 5.0]
    >>> _create_circle_skyregion_from_coords(coords)
    <CircleSkyRegion(center=<SkyCoord (ICRS): (ra, dec) in deg (10., 20.)>, radius=5.0 deg)>

    """
    sky_coordinates = SkyCoord(coords[0], coords[1], frame=frame, unit=unit)
    sky_region = CircleSkyRegion(center=sky_coordinates, radius=coords[2] * unit)

    return sky_region


def _create_ellipse_skyregion_from_coords(coords, frame='icrs', unit=u.deg):
    """Create a `regions.EllipseSkyRegion` from given coordinates.

    Parameters
    ----------
    coords : list
        List of coordinates in the form of [ra, dec, width, height, angle].

    frame : str, optional
        Coordinate frame. Default is 'icrs'.

    unit : `~astropy.units.Unit`, optional
        Unit of the coordinates. Default is `~astropy.units.deg`.

    Returns
    -------
    sky_region : `regions.EllipseSkyRegion`
        An ellipse sky region.

    Examples
    --------
    Create a `regions.EllipseSkyRegion` from given coordinates:

    >>> from jdaviz.core.region_translators import _create_ellipse_skyregion_from_coords
    >>> coords = [10.0, 20.0, 5.0, 3.0, 45.0]
    >>> _create_ellipse_skyregion_from_coords(coords)
    <EllipseSkyRegion(center=<SkyCoord (ICRS): (ra, dec) in deg
        (10., 20.)>, width=5.0 deg, height=3.0 deg, angle=45.0 deg)>

    """
    sky_coordinates = SkyCoord(coords[0], coords[1], frame=frame, unit=unit)
    sky_region = EllipseSkyRegion(center=sky_coordinates, width=coords[2] * unit,
                                  height=coords[3] * unit, angle=coords[4] * unit)

    return sky_region


SUPPORTED_STCS_SHAPE_VALUES = ('POLYGON', 'CIRCLE', 'ELLIPSE', )
SUPPORTED_STCS_FRAME_VALUES = ('ICRS', 'J2000', 'FK5', )
STCS_SHAPE_PATTERN = '|'.join(SUPPORTED_STCS_SHAPE_VALUES)
STCS_FRAME_PATTERN = '|'.join(SUPPORTED_STCS_FRAME_VALUES)
SUPPORTED_STCS_PATTERN = re.compile(
    fr"^(?P<shape>({'|'.join(SUPPORTED_STCS_SHAPE_VALUES)}))"
    fr"(?P<frame>(\s+({'|'.join(SUPPORTED_STCS_FRAME_VALUES)})))*"
    r"(?P<coordinates>(\s+-?\d+\.\d+){2,}$)",
    re.IGNORECASE)
SKY_REGION_FROM_COORDS_FACTORY = {
    'polygon': _create_polygon_skyregion_from_coords,
    'circle': _create_circle_skyregion_from_coords,
    'ellipse': _create_ellipse_skyregion_from_coords,
}


def is_stcs_string(stcs_string):
    """Check if the given string is a valid STC-S string.

    Parameters
    ----------
    stcs_string : str
        A string to check.

    Returns
    -------
    is_stcs : bool
        `True` if the given string is a valid STC-S string, otherwise `False`.

    Examples
    --------
    Check if a given string is a valid STC-S string:

    >>> from jdaviz.core.region_translators import is_stcs_string
    >>> stcs_string = 'POLYGON ICRS 10.0 20.0 30.0 40.0 50.0 60.0'
    >>> is_stcs_string(stcs_string)
    True

    """

    try:
        return bool(SUPPORTED_STCS_PATTERN.match(stcs_string))
    except Exception:
        return False


def stcs_string2region(stcs_string):
    """Convert a given STC-S string to ``regions`` shape.

    Parameters
    ----------
    stcs_string : str
        A valid STC-S string.

    Returns
    -------
    sky_region : `regions.Region`
        A supported ``regions`` shape.

    Raises
    ------
    ValueError
        Invalid inputs.

    Examples
    --------
    Translate an STC-S region to `regions.PolygonSkyRegion`:

    >>> from jdaviz.core.region_translators import stcs_string2region
    >>> stcs_string = 'POLYGON ICRS 10.0 20.0 30.0 40.0 50.0 60.0'
    >>> stcs_string2region(stcs_string)
    <PolygonSkyRegion(vertices=<SkyCoord (ICRS): (ra, dec) in deg
        [(10., 20.), (30., 40.), (50., 60.)]>)>

    """

    if not isinstance(stcs_string, str):
        raise ValueError('STC-S string must be a string.')

    _match = SUPPORTED_STCS_PATTERN.match(stcs_string)

    # Check if the STC-S string is valid, otherwise raise an error
    if not _match:
        raise ValueError(f'Invalid STC-S string: {stcs_string}')

    # Extract the shape, frame, and coordinates from the STC-S string
    shape = _match.group('shape').strip().lower()
    frame = _match.group('frame').strip().lower() if _match.group('frame') else 'icrs'
    coordinates = list(map(float, _match.group('coordinates').strip().split()))

    return SKY_REGION_FROM_COORDS_FACTORY[shape](coordinates, frame=frame)


def region2stcs_string(region):
    """Convert a `regions.Region` object to an STC-S string.

    Currently supports only `CircleSkyRegion` and `EllipseSkyRegion`.

    Parameters
    ----------
    region : `regions.Region`
        Sky region to convert to STC-S format.

    Returns
    -------
    stcs_str : str
        STC-S string representing the region, using the format:
        - `CIRCLE <FRAME> <RA> <DEC> <RADIUS>`
        - `ELLIPSE <FRAME> <RA> <DEC> <WIDTH> <HEIGHT> <ANGLE>`
    """
    try:
        shape, coords = SHAPE_COORDS_FROM_SKY_REGION_FACTORY[region.__class__](region)
    except KeyError:
        raise NotImplementedError(f"{region.__class__.__name__} is not supported for STC-S export")

    frame = region.center.frame.name.upper()
    if frame not in SUPPORTED_STCS_FRAME_VALUES:
        raise ValueError(f"Frame '{frame}' is not supported for STC-S export."
                         f"Supported: {SUPPORTED_STCS_FRAME_VALUES}")

    coord_parts = [f"{x:.6f}" for x in coords]

    return f"{shape} {frame} {' '.join(coord_parts)}"


def _shape_coords_from_circle_sky_region(region):
    return (
        "CIRCLE", [
            region.center.ra.deg,
            region.center.dec.deg,
            region.radius.to_value(u.deg),
        ],
    )


def _shape_coords_from_ellipse_sky_region(region):
    return (
        "ELLIPSE", [
            region.center.ra.deg,
            region.center.dec.deg,
            region.width.to_value(u.deg),
            region.height.to_value(u.deg),
            region.angle.to_value(u.deg),
        ],
    )


SHAPE_COORDS_FROM_SKY_REGION_FACTORY = {
    CircleSkyRegion: _shape_coords_from_circle_sky_region,
    EllipseSkyRegion: _shape_coords_from_ellipse_sky_region,
}


def positions2pixcoord(positions):
    """Convert ``photutils`` aperture positions to `~regions.PixCoord`
    that is acceptable by ``regions`` shape.
    """
    if positions.shape != (2, ):
        raise ValueError('regions shape only accepts scalar x and y positions')
    if isinstance(positions, u.Quantity):
        pixcoord = PixCoord(x=positions[0].value, y=positions[1].value)
    else:
        pixcoord = PixCoord(x=positions[0], y=positions[1])
    return pixcoord
